
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import os
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
import sys
import importlib
import re
from holado_core.common.tools.tools import Tools
from holado_core.common.tools.converters.converter import Converter
from holado_value.common.tables.value_table_manager import ValueTableManager
from holado_value.common.tools.value_types import ValueTypes
import inspect
from typing import NamedTuple
from holado_core.common.exceptions.holado_exception import HAException
from holado_python.standard_library.typing import Typing
from holado.common.handlers.undefined import undefined_argument, undefined_value

logger = logging.getLogger(__name__)

try:
    import google.protobuf.message
    import google.protobuf.descriptor
    from google.protobuf.descriptor import FieldDescriptor
    # import google.protobuf.pyext
    from google.protobuf.internal import api_implementation
    from holado_protobuf.ipc.protobuf.protobuf_converter import SortOrder
    
    # logger.info(f"Protobuf internal API implementation is of type '{api_implementation.Type()}'")
    if api_implementation.Type() == 'cpp':
        # from google.protobuf.pyext.cpp_message import GeneratedProtocolMessageType
        from google.protobuf.pyext._message import ScalarMapContainer as ScalarMap  # @UnresolvedImport @UnusedImport
        from google.protobuf.pyext._message import MessageMapContainer as MessageMap  # @UnresolvedImport @UnusedImport
        # from google.protobuf.pyext._message import RepeatedScalarContainer as RepeatedScalarFieldContainer
        # from google.protobuf.pyext._message import RepeatedCompositeContainer as RepeatedCompositeFieldContainer
    else:
        # from google.protobuf.internal.python_message import GeneratedProtocolMessageType
        from google.protobuf.internal.containers import ScalarMap  # @UnusedImport @Reimport
        from google.protobuf.internal.containers import MessageMap  # @UnusedImport @Reimport
        from google.protobuf.internal.containers import RepeatedScalarFieldContainer, RepeatedCompositeFieldContainer  # @UnusedImport
    
    with_protobuf = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"ProtobufMessages is not available. Initialization failed on error: {exc}")
    with_protobuf = False




class ProtobufMessages(object):
    
    @classmethod
    def is_available(cls):
        return with_protobuf

    if with_protobuf:
        
        @classmethod
        def is_descriptor_enum(cls, obj):
            return isinstance(obj, google.protobuf.descriptor.EnumDescriptor)
    
        @classmethod
        def is_descriptor_message(cls, obj):
            return isinstance(obj, google.protobuf.descriptor.Descriptor)
    
        @classmethod
        def is_object_message(cls, obj):
            return isinstance(obj, google.protobuf.message.Message)
    
        @classmethod
        def is_object_enum(cls, obj):
            class_name = type(obj).__name__.lower()
            return "enum" in class_name
    
        @classmethod
        def is_object_repeated(cls, obj):
            class_name = type(obj).__name__.lower()
            return "repeated" in class_name
    
        @classmethod
        def is_object_map(cls, obj):
            class_name = type(obj).__name__.lower()
            return "map" in class_name and hasattr(obj, 'items') and callable(obj.items)
            # return isinstance(obj, ScalarMap) or isinstance(obj, MessageMap)
            
        def is_message_field_set(self, obj, field_name):
            return self.__is_message_field_set(obj, field_name)
        
        def __init__(self): 
            self.__message_types_by_fullname = {}
            self.__enum_types_by_fullname = {}
            self.__enum_data_by_fullname = {}
            self.__regex_attribute_fullname = re.compile(r'^(.*?)(?:\[(.+)\])?$')
            
            self.__registered_types = []
        
        def initialize(self):
            self._register_types()
            
        def import_all_compiled_proto(self, compiled_proto_path, package_name=None, raise_if_not_exist=True):
            """Register a folder path containing compiled proto files. Usually it corresponds to the parameter '--python_out' passed to proto compiler."""
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[ProtobufMessages] Importing all compiled proto in '{compiled_proto_path}'...")
            
            if package_name is None:
                package_name = ""
            
            if os.path.exists(compiled_proto_path):
                if os.path.isfile(compiled_proto_path):
                    proto_path = os.path.dirname(compiled_proto_path)
                    sys.path.append(proto_path)
                    self.__import_compiled_proto(compiled_proto_path, package_name)
                elif os.path.isdir(compiled_proto_path):
                    sys.path.append(compiled_proto_path)
                    self.__import_all_compiled_proto(compiled_proto_path, package_name)
                else:
                    raise TechnicalException(f"Unmanaged path '{compiled_proto_path}'")
            else:
                msg = f"Path '{compiled_proto_path}' doesn't exist"
                if raise_if_not_exist:
                    raise TechnicalException(msg)
                else:
                    logger.warning(msg)
        
        def __import_all_compiled_proto(self, compiled_proto_path, package_name):
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"[ProtobufMessages] Importing all compiled proto in '{compiled_proto_path}' (package: '{package_name}')...")
            if os.path.isdir(compiled_proto_path):
                lp = os.listdir(compiled_proto_path)
                for cp in lp:
                    if not cp.startswith((".", "_")):
                        cur_proto_path = os.path.join(compiled_proto_path, cp)
                        
                        if os.path.isfile(cur_proto_path):
                            self.__import_compiled_proto(cur_proto_path, package_name)
                        elif os.path.isdir(cur_proto_path):
                            cur_package_name = f"{package_name}.{cp}" if package_name is not None and len(package_name) > 0 else cp
                            self.__import_all_compiled_proto(cur_proto_path, cur_package_name)
                        else:
                            raise TechnicalException(f"Unmanaged path '{cur_proto_path}'")
            else:
                raise TechnicalException(f"Unmanaged path '{compiled_proto_path}'")
        
        def __import_compiled_proto(self, compiled_proto_file_path, package_name):
            if not os.path.isfile(compiled_proto_file_path):
                raise TechnicalException(f"Compiled proto path '{compiled_proto_file_path}' is not a file")
            if not compiled_proto_file_path.endswith("_pb2.py"):
                return
    
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
    
                logger.trace(f"[ProtobufMessages] Importing compiled proto file '{compiled_proto_file_path}' (package: '{package_name}')...")
            
            filename = os.path.splitext(os.path.basename(compiled_proto_file_path))[0]
            module_name = f"{package_name}.{filename}" if package_name is not None and len(package_name) > 0 else filename
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"[ProtobufMessages] Importing module '{module_name}'")
            module = importlib.import_module(module_name)
            
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            
                logger.trace(f"[ProtobufMessages] DESCRIPTOR of module '{module_name}': {self.__represent_descriptor(module.DESCRIPTOR)}")
            module_package = module.DESCRIPTOR.package if hasattr(module.DESCRIPTOR, 'package') else package_name
            self.__import_compiled_proto_object(module, module_package)
    
        def __import_compiled_proto_object(self, module_or_object, module_or_object_fullname):
            # Import message types
            if hasattr(module_or_object.DESCRIPTOR, 'message_types_by_name'):
                for mt_name in module_or_object.DESCRIPTOR.message_types_by_name:
                    self.__import_compiled_proto_message(module_or_object, module_or_object_fullname, mt_name)
                
            # Import enum types
            if hasattr(module_or_object.DESCRIPTOR, 'enum_types_by_name'):
                for et_name in module_or_object.DESCRIPTOR.enum_types_by_name:
                    self.__import_compiled_proto_enum(module_or_object, module_or_object_fullname, et_name)
                
            # Import nested types
            if hasattr(module_or_object.DESCRIPTOR, 'nested_types_by_name'):
                for nt_name in module_or_object.DESCRIPTOR.nested_types_by_name:
                    nt = module_or_object.DESCRIPTOR.nested_types_by_name[nt_name]
                    if self.is_descriptor_message(nt):
                        self.__import_compiled_proto_message(module_or_object, module_or_object_fullname, nt_name)
                    elif self.is_descriptor_enum(nt):
                        self.__import_compiled_proto_enum(module_or_object, module_or_object_fullname, nt_name)
                    else:
                        raise TechnicalException(f"Unmanaged nested type '{nt_name}' having descriptor: {self.__represent_descriptor(nt)}")
    
        def __import_compiled_proto_message(self, module_or_object, module_or_object_fullname, message_type_name):
            if hasattr(module_or_object, message_type_name):
                mt = getattr(module_or_object, message_type_name)
                mt_fullname = f"{module_or_object_fullname}.{message_type_name}" if module_or_object_fullname is not None and len(module_or_object_fullname) > 0 else message_type_name
                self.__message_types_by_fullname[mt_fullname] = mt
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"[ProtobufMessages] New managed message type '{mt_fullname}' (type: '{mt.__qualname__}')")
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"[ProtobufMessages] Message type '{mt_fullname}': {Tools.represent_object(mt)}")
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"[ProtobufMessages] DESCRIPTOR of message type '{mt_fullname}': {self.__represent_descriptor(mt.DESCRIPTOR)}")
                
                # Import nested types
                self.__import_compiled_proto_object(mt, mt_fullname)
            else:
                raise TechnicalException(f"Not found message type '{message_type_name}' in '{module_or_object_fullname}': {Tools.represent_object(module_or_object)}")
    
        def __import_compiled_proto_enum(self, module_or_object, module_or_object_fullname, enum_type_name):
            if hasattr(module_or_object, enum_type_name):
                et = getattr(module_or_object, enum_type_name)
                et_fullname = f"{module_or_object_fullname}.{enum_type_name}" if module_or_object_fullname is not None and len(module_or_object_fullname) > 0 else enum_type_name
                self.__enum_types_by_fullname[et_fullname] = et
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"[ProtobufMessages] New managed enum type '{et_fullname}' (type: '{Typing.get_object_class_fullname(et)}')")
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"[ProtobufMessages] Enum type '{et_fullname}': {Tools.represent_object(et)}")
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"[ProtobufMessages] DESCRIPTOR of enum type '{et_fullname}': {self.__represent_descriptor(et.DESCRIPTOR)}")
            else:
                raise TechnicalException(f"Not found enum type '{enum_type_name}' in '{module_or_object_fullname}': {Tools.represent_object(module_or_object)}")
    
        def has_object_type(self, type_fullname):
            """Return if type fullname is known."""
            return self.has_message_type(type_fullname) or self.has_enum_type(type_fullname)
    
        def has_message_type(self, message_type_fullname):
            """Return if message type fullname is known."""
            return message_type_fullname in self.__message_types_by_fullname
    
        def has_enum_type(self, enum_type_fullname):
            """Return if enum type fullname is known."""
            return enum_type_fullname in self.__enum_types_by_fullname
    
        def get_object_type(self, type_fullname):
            """Return type object for given type fullname."""
            if self.has_message_type(type_fullname):
                return self.get_message_type(type_fullname)
            else:
                raise FunctionalException(f"Unknown type fullname '{type_fullname}'")
    
        def get_message_type(self, message_type_fullname):
            """Return type object for given message type fullname."""
            if self.has_message_type(message_type_fullname):
                return self.__message_types_by_fullname[message_type_fullname]
            else:
                raise FunctionalException(f"Unknown message type fullname '{message_type_fullname}'")
    
        def get_enum_type(self, enum_type_fullname):
            """Return type object for given message type fullname."""
            if self.has_enum_type(enum_type_fullname):
                return self.__enum_types_by_fullname[enum_type_fullname]
            else:
                raise FunctionalException(f"Unknown enum type fullname '{enum_type_fullname}' (known enums: {list(self.__enum_types_by_fullname.keys())})")
            
        def get_enum_name(self, enum_value, enum_type_fullname):
            enum_type = self.get_enum_type(enum_type_fullname)
            return self.__get_enum_name(enum_value, enum_type=enum_type)
    
        def get_enum_value(self, fullname=None, enum_type_fullname=None, enum_name=None):
            if fullname is not None:
                enum_type_fullname, enum_name = fullname.rsplit('.', 1)
            enum_type = self.get_enum_type(enum_type_fullname)
            return self.__get_enum_value(enum_name, enum_type=enum_type)
    
        def new_object(self, type_fullname):
            """Return a new object of given type fullname."""
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"Creating new object of type '{type_fullname}'...")
            if self.has_object_type(type_fullname):
                if self.has_message_type(type_fullname):
                    mt = self.get_message_type(type_fullname)
                    res = mt()
                else:
                    raise TechnicalException(f"Unmanaged object creation for type fullname '{type_fullname}' (all managed types are logged at start with level DEBUG)")
            else:
                raise TechnicalException(f"Unknown type fullname '{type_fullname}' (all known types are logged at start with level DEBUG)")
            
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"New object of type '{type_fullname}' => {res}")
            return res
    
        def new_message(self, message_type_fullname, fields_table=None, fields_dict=None, serialized_string=None):
            """Return a new message of given message type fullname.
            The content of the message can be filled from serialized string, or from field values in a Name/Value table.
            """
            res = self.new_object(message_type_fullname)
            
            if serialized_string is not None:
                res.ParseFromString(serialized_string)
            elif fields_table is not None:
                if ValueTableManager.verify_table_is_name_value_table(fields_table, raise_exception=False):
                    self.__set_object_fields_with_name_value_table(res, fields_table)
                else:
                    raise TechnicalException(f"When defining parameter fields_table, it must be a Name/Value table: [{Typing.get_object_class_fullname(fields_table)}]\n{fields_table.represent(4)}")
            elif fields_dict is not None:
                self.__set_object_fields_with_dict(res, fields_dict)
            
            if Tools.do_log(logger, logging.DEBUG):
                # logger.debug(f"New message of type '{message_type_fullname}' => {res}")
                logger.debug(f"New message of type '{message_type_fullname}' => {Tools.represent_object(res)}")
            return res
            
        def get_object_field_names(self, obj, recursive=False, uncollapse_repeated=False, add_repeated_index=False, with_unset=True, sort_order=SortOrder.Definition):
            res = self.__get_object_field_names(obj, recursive=recursive, uncollapse_repeated=uncollapse_repeated, add_repeated_index=add_repeated_index, with_unset=with_unset, is_message_field=False, sort_order=sort_order)
            # logger.trace(f"Object of type '{self.get_object_type_fullname(obj)}' has field names: {res}")
            return res
            
        def get_message_field_names(self, obj, recursive=False, uncollapse_repeated=False, add_repeated_index=False, with_unset=True):
            res = self.__get_message_field_names(obj, recursive=recursive, uncollapse_repeated=uncollapse_repeated, add_repeated_index=add_repeated_index, with_unset=with_unset)
            # logger.trace(f"Message of type '{self.get_object_type_fullname(obj)}' has field names: {res}")
            return res
            
        def get_object_field_values(self, obj, attribute_expression):
            """
            Return a list of values for attribute names matching given expression.
            Attribute name expression can contain "." for sub-fields, and "[]" to access all elements of the repeated field.
            The list contains only one value unless a repeated field is included in attribute expression with suffix '[]'.
            """ 
            attr_names = attribute_expression.split('.')
            res_list = [(obj, "")]
            for attr_name in attr_names:
                old_list = res_list
                if attr_name.endswith("[]"):
                    if len(old_list) > 1:
                        raise TechnicalException("Uncollapse two different repeated fields in the same message is not managed.")
                    cur_res = old_list[0][0]
                    cur_fullname = old_list[0][1]
                    real_attr_name = attr_name[:-2]
                    
                    cur_attr_fullname = cur_fullname + '.' + real_attr_name if len(cur_fullname) > 0 else real_attr_name
                    if not hasattr(cur_res, real_attr_name):
                        raise TechnicalException(f"Attribute '{cur_attr_fullname}' doesn't exist in object [{cur_res}]")
                    
                    res_list = []
                    for index, attr_obj in enumerate(getattr(cur_res, real_attr_name)):
                        new_attr_fullname = f"{cur_attr_fullname}[{index}]"
                        res_list.append((attr_obj, new_attr_fullname))
                else:
                    res_list = []
                    for obj, obj_attr_fullname in old_list:
                        new_attr_fullname = f"{obj_attr_fullname}.{attr_name}"
                        if self.__has_object_field(obj, field_name=attr_name):
                            attr_obj = self.__get_object_field(obj, field_name=attr_name)
                            res_list.append((attr_obj, new_attr_fullname))
                        else:
                            raise TechnicalException(f"Attribute '{new_attr_fullname}' doesn't exist in object [{obj}]")
                        
            return [obj for obj, _ in res_list]
                
        def has_object_field(self, obj, attribute_expression, create_field=False):
            leaf_obj, _, attr_last_name = self.__get_leaf_object_and_attribute_names(obj, attribute_expression, create_field=create_field)
            return self.__has_object_field(leaf_obj, field_fullname=attr_last_name)
            
        def get_object_field_value(self, obj, attribute_expression, create_field=False):
            """
            Return the value for attribute name matching given expression.
            Attribute name expression can contain "." for sub-fields.
            """ 
            attr_names = attribute_expression.split('.')
            res = obj
            attr_fullname = ""
            for attr_name in attr_names:
                if self.__has_object_field(res, field_fullname=attr_name):
                    res = self.__get_object_field(res, field_fullname=attr_name, create_field=create_field)
                    attr_fullname = attr_fullname + '.' + attr_name if len(attr_fullname) > 0 else attr_name
                else:
                    attr_name, _ = self.__split_attribute_fullname(attr_name)
                    if len(attr_fullname) > 0:
                        attr_fullname = attr_fullname + '.' + attr_name
                        raise FunctionalException(f"Field '{attr_fullname}' doesn't exist in message type '{self.get_object_type_fullname(obj)}' (in sub-message type '{self.get_object_type_fullname(res)}', existing fields: {self.__get_object_descriptor_field_names(res, False)})")
                    else:
                        raise FunctionalException(f"Field '{attr_name}' doesn't exist in message type '{self.get_object_type_fullname(obj)}' (existing fields: {self.__get_object_descriptor_field_names(obj, False)})")
                
            return res
                
        def set_object_field_value(self, obj, attribute_expression, value):
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Setting protobuf object ({id(obj)}) field '{attribute_expression}' with value [{value}] (type: {Typing.get_object_class_fullname(value)})")
            leaf_obj, leaf_attribute_expression, attr_last_name = self.__get_leaf_object_and_attribute_names(obj, attribute_expression, create_field=True)
                
            # Set value
            if self.__has_object_field(leaf_obj, field_fullname=attr_last_name):
                self.__set_object_field(leaf_obj, field_fullname=attr_last_name, value=value, create_field=True)
            else:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"Field '{attr_last_name}' doesn't exist in type '{self.get_object_type_fullname(leaf_obj)}'")
                attr_name, _ = self.__split_attribute_fullname(attr_last_name)
                if leaf_attribute_expression is not None:
                    attr_fullname = leaf_attribute_expression + '.' + attr_name
                    raise FunctionalException(f"Field '{attr_fullname}' doesn't exist in object type '{self.get_object_type_fullname(obj)}' (in sub-message type '{self.get_object_type_fullname(leaf_obj)}', existing fields: {self.__get_object_descriptor_field_names(leaf_obj, False)})")
                else:
                    raise FunctionalException(f"Field '{attr_name}' doesn't exist in object type '{self.get_object_type_fullname(obj)}' (existing fields: {self.__get_object_descriptor_field_names(obj, False)})")
                
        def __get_leaf_object_and_attribute_names(self, obj, attribute_expression, create_field=False):
            try:
                leaf_attribute_expression, attr_last_name = attribute_expression.rsplit('.', 1)
            except ValueError:
                leaf_obj = obj
                leaf_attribute_expression = None
                attr_last_name = attribute_expression
            else:
                # Get "leaf" object (in object tree) containing the leaf attribute to set
                leaf_obj = self.get_object_field_value(obj, leaf_attribute_expression, create_field=create_field)
            return leaf_obj, leaf_attribute_expression, attr_last_name
    
        def __get_object_field_names(self, obj, recursive=False, uncollapse_repeated=False, add_repeated_index=True, with_unset=True, prefix="", is_message_field=False, sort_order=SortOrder.Definition):
            res = []
            if uncollapse_repeated and ProtobufMessages.is_object_repeated(obj):
                if add_repeated_index:
                    for index, value in enumerate(obj):
                        new_prefix = f"{prefix}[{index}]"
                        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                            logger.trace(f"Adding field names of repeated with prefix '{new_prefix}' (type: {Typing.get_object_class_fullname(value)})")
                        res.extend(self.__get_object_field_names(value, recursive=recursive, uncollapse_repeated=uncollapse_repeated, add_repeated_index=add_repeated_index, with_unset=with_unset, prefix=new_prefix, is_message_field=is_message_field, sort_order=sort_order))
                else:
                    new_prefix = f"{prefix}[]"
                    value = obj[0]
                    if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"Adding field names of repeated with prefix '{new_prefix}' (type: {Typing.get_object_class_fullname(value)})")
                    res.extend(self.__get_object_field_names(value, recursive=recursive, uncollapse_repeated=uncollapse_repeated, add_repeated_index=add_repeated_index, with_unset=with_unset, prefix=new_prefix, is_message_field=is_message_field, sort_order=sort_order))
            elif (recursive or not is_message_field) and ProtobufMessages.is_object_map(obj):
                sorted_dict = dict(sorted(obj.items()))
                for key, value in sorted_dict.items():
                    key_prefix = f"{prefix}[{key}]" if len(prefix) > 0 else key
                    if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"Adding field names of object of prefix '{key_prefix}' (type: {Typing.get_object_class_fullname(value)})")
                    res.extend(self.__get_object_field_names(value, recursive=recursive, uncollapse_repeated=uncollapse_repeated, add_repeated_index=add_repeated_index, with_unset=with_unset, prefix=key_prefix, is_message_field=is_message_field))
            elif (recursive or not is_message_field) and ProtobufMessages.is_object_message(obj):
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"Adding field names of message of prefix '{prefix}' (type: {Typing.get_object_class_fullname(obj)})")
                res.extend(self.__get_message_field_names(obj, recursive=recursive, uncollapse_repeated=uncollapse_repeated, add_repeated_index=add_repeated_index, with_unset=with_unset, prefix=prefix, sort_order=sort_order))
            elif len(prefix) > 0:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"Adding field name '{prefix}' (value type: {Typing.get_object_class_fullname(obj)})")
                res.append(prefix)
            else:
                # logger.trace(f"Adding field name '{attr_name}' (value type: {Typing.get_object_class_fullname(attr_val)})")
                raise TechnicalException(f"Object has no field and prefix is empty (object: {obj} ; type: {Typing.get_object_class_fullname(obj)})")
            return res
    
        def __get_message_field_names(self, obj, recursive=False, uncollapse_repeated=False, add_repeated_index=True, with_unset=True, prefix="", sort_order=SortOrder.Definition):
            res = []
            attribute_names = self.__get_object_descriptor_field_names(obj, sort_order=sort_order)
            for attr_name in attribute_names:
                attr_val = getattr(obj, attr_name)
                new_prefix = f"{prefix + '.' if len(prefix) > 0 else ''}{attr_name}"
                
                # Skip field not set
                if not with_unset:
                    set_status = self.__is_message_field_set(obj, attr_name)
                    if set_status == False:
                        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                            logger.trace(f"Hide unset field '{attr_name}' in object type '{self.get_object_type_fullname(obj)}' (field type: {Typing.get_object_class_fullname(attr_val)})")
                        continue
                
                # Skip optional fields that are not set
                if self.__is_message_field_optional(obj, attr_name):
                    set_status = self.__is_message_field_set(obj, attr_name)
                    if set_status == False:
                        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                            logger.trace(f"Hide unset optional field '{attr_name}' in object type '{self.get_object_type_fullname(obj)}' (field type: {Typing.get_object_class_fullname(attr_val)})")
                        continue
                
                # Skip oneof field that is not set
                if self.__is_message_field_oneof_field(obj, attr_name):
                    set_status = self.__is_message_field_set(obj, attr_name)
                    if set_status == False:
                        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                            logger.trace(f"Hide unset oneof field '{attr_name}' in object type '{self.get_object_type_fullname(obj)}' (field type: {Typing.get_object_class_fullname(attr_val)})")
                        continue
                
                res.extend(self.__get_object_field_names(attr_val, recursive=recursive, uncollapse_repeated=uncollapse_repeated, add_repeated_index=add_repeated_index, with_unset=with_unset, prefix=new_prefix, is_message_field=True, sort_order=sort_order))
            return res
        
        def __is_message_field_oneof_field(self, obj, field_name):
            field_descr = self.__get_object_field_descriptor(obj, field_name)
            res = (field_descr.label == FieldDescriptor.LABEL_OPTIONAL 
                   and field_descr.containing_oneof is not None
                   and field_descr.has_presence)
            return res
        
        def __is_message_field_optional(self, obj, field_name):
            field_descr = self.__get_object_field_descriptor(obj, field_name)
            res = (field_descr.label == FieldDescriptor.LABEL_OPTIONAL 
                   and field_descr.containing_oneof is not None and len(field_descr.containing_oneof.fields) == 1
                   and field_descr.has_presence)
            return res
        
        def __is_message_field_repeated(self, obj, field_name):
            field_descr = self.__get_object_field_descriptor(obj, field_name)
            return (field_descr.label == FieldDescriptor.LABEL_REPEATED)
        
        def __is_message_field_required(self, obj, field_name):
            field_descr = self.__get_object_field_descriptor(obj, field_name)
            return (field_descr.label == FieldDescriptor.LABEL_REQUIRED)
        
        def __is_message_field_set(self, obj, field_name):
            """
            If field can distinguish between unpopulated and default values, return if field is set, else return None.
            """
            try:
                return obj.HasField(field_name)
            except ValueError:
                # logger.trace(f"Field '{field_name}' is not optional. Got error: {exc}")
                return None
            
        def __get_object_descriptor_field_names(self, obj, raise_exception=True, sort_order=SortOrder.Definition):
            #TODO EKL: manage oneof fields => return only the name of the oneof field that is defined 
            if hasattr(obj, 'DESCRIPTOR'):
                descriptor = getattr(obj, 'DESCRIPTOR')
                # TODO: When it will be possible with python generated code, remove from result the deprecated fields
                # logger.info(f"+++++++++++++ descriptor: {self.__represent_descriptor(descriptor)}")
                res = [f.name for f in descriptor.fields if not hasattr(f, "isDeprecated")]
                if sort_order == SortOrder.Alphabetic:
                    res.sort()
                return res
            else:
                # return Typing.get_object_attribute_names(obj)
                if raise_exception:
                    raise TechnicalException(f"Not found attribute 'DESCRIPTOR' in object of type '{self.get_object_type_fullname(obj)}' [{obj}]")
                else:
                    return []
                
        def __get_object_field_descriptor(self, obj, field_name):
            if hasattr(obj, 'DESCRIPTOR'):
                descriptor = getattr(obj, 'DESCRIPTOR')
                if field_name in descriptor.fields_by_name:
                    return descriptor.fields_by_name[field_name]
                else:
                    raise FunctionalException(f"Field '{field_name}' doesn't exist in type '{self.get_object_type_fullname(obj)}'")
            else:
                raise TechnicalException(f"Not found attribute 'DESCRIPTOR' in object of type '{self.get_object_type_fullname(obj)}' [{obj}]")
                
        def __has_object_field(self, obj, field_name=None, field_fullname=None):
            param_in_brackets = None
            if field_name is None:
                field_name, param_in_brackets = self.__split_attribute_fullname(field_fullname)
                
            if len(field_name) == 0:
                # Manage field_fullname in format "[XXX]"
                if param_in_brackets is not None:
                    if self.is_object_repeated(obj):
                        if Converter.is_integer(param_in_brackets):
                            li_index = int(param_in_brackets)
                            res = li_index < len(obj)
                        else:
                            raise FunctionalException(f"For repeated objects, the parameter in brackets must be an integer (field fullname: '{field_fullname}')")
                    elif self.is_object_map(obj):
                        res = param_in_brackets in obj
                        # if not res:
                        #     logger.trace(f"++++++ Key '{param_in_brackets}' in not in map {obj}")
                    else:
                        raise TechnicalException(f"Unexpected brackets in field fullname '{field_fullname}' for object of type '{self.get_object_type_fullname(obj)}'")
                else:
                    raise TechnicalException(f"Unexpected field " + f"fullname '{field_fullname}'" if field_fullname is not None else f"name '{field_name}'")
            elif self.is_object_map(obj):
                res = field_name in obj
            else:
                res = field_name in self.__get_object_descriptor_field_names(obj, False)
                
            if not res:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"Field '{field_name}' doesn't exist in type '{self.get_object_type_fullname(obj)}' (existing fields: {self.__get_object_descriptor_field_names(obj, False)})")
            return res
        
        def __get_object_field_type_fullname(self, obj, field_name):
            if not self.__has_object_field(obj, field_name):
                raise FunctionalException()
            
            field_descr = self.__get_object_field_descriptor(obj, field_name)
            if field_descr.message_type:
                return self.get_object_type_fullname(descriptor=field_descr.message_type)
            elif field_descr.enum_type:
                return self.get_object_type_fullname(descriptor=field_descr.enum_type)
            else:
                return field_descr.type
        
        def __get_object_field(self, obj, field_name=None, field_fullname=None, create_field=False):
            param_in_brackets = None
            if field_name is None:
                field_name, param_in_brackets = self.__split_attribute_fullname(field_fullname)
            
            if len(field_name) == 0:
                # Manage field_fullname in format "[XXX]"
                if param_in_brackets is not None:
                    if self.is_object_repeated(obj):
                        if Converter.is_integer(param_in_brackets):
                            li_index = int(param_in_brackets)
                            res = self.__get_object_repeated_by_index(obj, li_index, add_index=create_field)
                        else:
                            raise FunctionalException(f"For repeated objects, the parameter in brackets must be an integer (field fullname: '{field_fullname}')")
                    elif self.is_object_map(obj):
                        res = obj[param_in_brackets]
                    else:
                        raise TechnicalException(f"Unexpected brackets in field fullname '{field_fullname}' for object of type '{self.get_object_type_fullname(obj)}'")
                else:
                    raise TechnicalException(f"Unexpected field " + f"fullname '{field_fullname}'" if field_fullname is not None else f"name '{field_name}'")
            
            elif self.is_object_map(obj):
                res = obj[field_name]
                
            elif hasattr(obj, field_name):
                if param_in_brackets is not None:
                    if Converter.is_integer(param_in_brackets):
                        li_index = int(param_in_brackets)
                        res = self.__get_object_repeated_field_by_index(obj, field_name, li_index, add_index=create_field)
                    else:
                        res = getattr(obj, field_name)[param_in_brackets]
                else:
                    res = getattr(obj, field_name)
                
                # Manage enum
                res = self.__get_enum_name_if_field_is_enum(res, obj=obj, field_name=field_name)
                
            elif self.__has_object_field(obj, field_name=field_name):
                if create_field:
                    field_type_fullname = self.__get_object_field_type_fullname(obj, field_name)
                    if isinstance(field_type_fullname, int):
                        raise TechnicalException("Unexpected case: the native types are expected to exist in object")
                    res = self.new_object(field_type_fullname)
                    setattr(obj, field_name, res)
                else:
                    raise FunctionalException(f"Field '{field_name}' exists in type '{self.get_object_type_fullname(obj)}' but not in instance [{obj}]")
            else:
                raise FunctionalException(f"Field '{field_name}' doesn't exist in type '{self.get_object_type_fullname(obj)}'")
    
            return res
        
        def __get_object_repeated_by_index(self, obj, index, add_index=False):
            if not ProtobufMessages.is_object_repeated(obj):
                raise FunctionalException(f"Object of type '{self.get_object_type_fullname(obj)}' is not a repeated")
            
            # Add repeated element if it doesn't exist
            if len(obj) < index + 1:
                if add_index:
                    for _ in range(index + 1 - len(obj)):
                        res = obj.add()
                else:
                    raise FunctionalException(f"Index {index} exceeds repeated length {len(obj)}")
            else:
                res = obj[index]
                
            return res
        
        def __get_object_repeated_field_by_index(self, obj, field_name, index, add_index=False):
            field_obj = getattr(obj, field_name)
            if not ProtobufMessages.is_object_repeated(field_obj):
                raise FunctionalException(f"Field '{field_name}' is not a repeated in type '{self.get_object_type_fullname(obj)}'")
            if len(field_obj) < index + 1 and not add_index:
                raise FunctionalException(f"Index {index} exceeds repeated length {len(field_obj)} (field '{field_name}' in object of type '{self.get_object_type_fullname(obj)}' [{obj}])")
            
            return self.__get_object_repeated_by_index(field_obj, index, add_index)
        
        def get_object_type_fullname(self, obj=None, obj_class=None, descriptor=None):
            # None type
            if obj is None and obj_class is None and descriptor is None:
                return None.__class__.__name__
    
            # Object -> object class
            if obj is not None:
                return self.get_object_type_fullname(obj_class=type(obj))
    
            # Object class -> object class descriptor
            if hasattr(obj_class, 'DESCRIPTOR'):
                return self.get_object_type_fullname(descriptor=obj_class.DESCRIPTOR)
            
            # Extract information from descriptor
            if descriptor:
                if hasattr(descriptor, 'full_name'):
                    return descriptor.full_name
                elif descriptor.containing_type is not None:
                    containing_fullname = self.get_object_type_fullname(obj_class=descriptor.containing_type)
                    return containing_fullname + '.' + descriptor.name
                else:
                    raise TechnicalException(f"Failed to extract type fullname from descriptor: {self.__represent_descriptor(descriptor)}")
                
            # Extract information from classzhDM
            if obj_class.__module__.endswith('_pb2'):
                package_name = os.path.splitext(obj_class.__module__)[0]
                return package_name + '.' + obj_class.__name__
            else:
                return obj_class.__module__ + '.' + obj_class.__name__
    
        def __set_object_fields_with_name_value_table(self, obj, table):
            # Verify table structure
            ValueTableManager.verify_table_is_name_value_table(table)
            
            for row in table.rows:
                if row.get_cell(1).value_type not in [ValueTypes.NotApplicable]:
                    if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"Setting protobuf object field with row ({row})")
                    
                    name = row.get_cell(0).value
                    value = row.get_cell(1).value
                    
                    self.set_object_field_value(obj, name, value)
                    
        def __set_object_fields_with_dict(self, obj, values_dict):
            for name, value in values_dict.items():
                self.set_object_field_value(obj, name, value)
        
        def __check_field_value_to_set_is_valid(self, value):
            if value is undefined_argument or value is undefined_value:
                raise ValueError(f"Special objects undefined_* are not valid field values")
            
        def __set_object_field(self, obj, field_name=None, field_fullname=None, value=None, create_field=False):
            in_brackets_str = None
            if field_name is None:
                field_name, in_brackets_str = self.__split_attribute_fullname(field_fullname)
            if not self.__has_object_field(obj, field_name):
                raise TechnicalException(f"Field '{field_name}' doesn't exist in type '{self.get_object_type_fullname(obj)}'")
            self.__check_field_value_to_set_is_valid(value)
            
            try:
                if hasattr(obj, field_name):
                    if in_brackets_str is not None:
                        from holado_test.scenario.step_tools import StepTools
                        in_brackets_obj = StepTools.evaluate_scenario_parameter(in_brackets_str)
                        if isinstance(in_brackets_obj, int):
                            li_index = in_brackets_obj
                            self.__set_object_repeated_field_by_index(obj, field_name, li_index, value, add_index=create_field)
                        else:
                            field_obj = getattr(obj, field_name)
                            field_obj[in_brackets_obj] = value
                    elif self.is_object_map(getattr(obj, field_name)):
                        field_obj = getattr(obj, field_name)
                        for key in value:
                            field_obj[key] = value[key]
                    elif self.is_object_repeated(getattr(obj, field_name)):
                        self.__set_object_repeated_field(obj, field_name, value, add_index=True)
                    else:
                        field_type_fullname = self.__get_object_field_type_fullname(obj, field_name)
                        if isinstance(field_type_fullname, int):
                            self.__set_object_field_attr(obj, field_name, value)
                        else:
                            field_descr = self.__get_object_field_descriptor(obj, field_name)
                            # logger.trace(f"+++++ Setting field '{field_name}' of descriptor: {self.__represent_descriptor(field_descr)}")
                            if field_descr.enum_type is not None:
                                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                                    logger.trace(f"Setting enum field '{field_name}' (type: {field_type_fullname}) with value [{value}] (type: {Typing.get_object_class_fullname(value)})")
                                enum_value = self.__get_enum_value(value, enum_type_descriptor=field_descr.enum_type)
                                self.__set_object_field_attr(obj, field_name, enum_value)
                            else:
                                field_obj = getattr(obj, field_name)
                                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                                    logger.trace(f"Setting field '{field_name}' (type: {field_type_fullname}) with value [{value}] (type: {Typing.get_object_class_fullname(value)})")
                                try:
                                    self.__set_object_value(field_obj, value)
                                except HAException as exc:
                                    raise TechnicalException(f"Unmanaged set of value [{value}] (type: {Typing.get_object_class_fullname(value)}) in field of type '{field_type_fullname}' (field '{field_name}' in object of type '{self.get_object_type_fullname(obj)}' [{obj}])\n  -> {exc.message}") from exc
                                except Exception as exc:
                                    raise TechnicalException(f"Unmanaged set of value [{value}] (type: {Typing.get_object_class_fullname(value)}) in field of type '{field_type_fullname}' (field '{field_name}' in object of type '{self.get_object_type_fullname(obj)}' [{obj}])\n  -> {str(exc)}") from exc
                elif self.__has_object_field(obj, field_name=field_name):
                    if create_field:
                        field_type_fullname = self.__get_object_field_type_fullname(obj, field_name)
                        if isinstance(field_type_fullname, int):
                            self.__set_object_field_attr(obj, field_name, value)
                        else:
                            res = self.new_object(field_type_fullname)
                            try:
                                self.__set_object_value(res, value)
                            except Exception as exc:
                                raise TechnicalException(f"Unmanaged set of value [{value}] (type: {Typing.get_object_class_fullname(value)}) in field of type '{field_type_fullname}' (field '{field_name}' in object of type '{self.get_object_type_fullname(obj)}' [{obj}])\n  -> {exc.message}") from exc
                            setattr(obj, field_name, res)
                    else:
                        raise FunctionalException(f"Field '{field_name}' exists in type '{self.get_object_type_fullname(obj)}' but not in instance [{obj}]")
                else:
                    raise FunctionalException(f"Field '{field_name}' doesn't exist in type '{self.get_object_type_fullname(obj)}'")
            except (FunctionalException, TechnicalException):
                raise
            except Exception as exc:
                if hasattr(obj, field_name):
                    field_obj = getattr(obj, field_name)
                    raise TechnicalException(f"Failed to set field '{field_name}' (type: '{Typing.get_object_class_fullname(field_obj)}') with value [{value}] (type '{Typing.get_object_class_fullname(value)}')") from exc
                else:
                    raise TechnicalException(f"Failed to set field '{field_name}' (type: Unknown) with value [{value}] (type '{Typing.get_object_class_fullname(value)}')") from exc
                    
        def __set_object_field_attr(self, obj, field_name, value):
            if value is None:
                obj.ClearField(field_name)
            else:
                setattr(obj, field_name, value)
    
        def __get_enum_name_if_field_is_enum(self, value, field_descriptor=None, obj=None, field_name=None):
            if field_descriptor is None:
                field_descriptor = self.__get_object_field_descriptor(obj, field_name)
            if field_descriptor.enum_type is not None:
                enum_name = self.__get_enum_name(value, enum_type_descriptor=field_descriptor.enum_type)
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"Getting enum field '{field_descriptor.name}' (type: {field_descriptor.enum_type} ; full_name: {field_descriptor.full_name}): value [{value}] -> name [{enum_name}]")
                return enum_name
            else:
                return value
    
        def __get_enum_name(self, name_or_value, enum_type=None, enum_type_descriptor=None):
            if enum_type is None and enum_type_descriptor is None:
                raise TechnicalException(f"Parameter 'enum_type' or 'enum_type_descriptor' must be defined")
            if enum_type is not None:
                enum_type_descriptor = enum_type.DESCRIPTOR
                
            if isinstance(name_or_value, str):
                return name_or_value
            elif isinstance(name_or_value, int):
                if name_or_value in enum_type_descriptor.values_by_number:
                    return self.__get_enum_name_for_value(enum_type_descriptor, name_or_value)
                else:
                    raise FunctionalException(f"Enum type '{enum_type_descriptor.full_name}' has no value '{name_or_value}' (possible values: {[k for k in enum_type_descriptor.values_by_number]})")
            elif self.is_object_repeated(name_or_value):
                return [self.__get_enum_name(v, enum_type_descriptor=enum_type_descriptor) for v in name_or_value]
            else:
                raise TechnicalException(f"Unexpected value type '{Typing.get_object_class_fullname(name_or_value)}'")
        
        def __get_enum_name_for_value(self, enum_type_descriptor, value):
            enum_type_fullname = self.get_object_type_fullname(descriptor=enum_type_descriptor)
            if not enum_type_fullname in self.__enum_data_by_fullname:
                self.__load_enum_data(enum_type_fullname, enum_type_descriptor)
            return self.__enum_data_by_fullname[enum_type_fullname].values_by_number[value].name
            
        def __get_enum_value(self, name_or_value, enum_type=None, enum_type_descriptor=None):
            if enum_type is None and enum_type_descriptor is None:
                raise TechnicalException(f"Parameter 'enum_type' or 'enum_type_descriptor' must be defined")
            if enum_type is not None:
                enum_type_descriptor = enum_type.DESCRIPTOR
                
            if isinstance(name_or_value, str):
                if name_or_value in enum_type_descriptor.values_by_name:
                    return enum_type_descriptor.values_by_name[name_or_value].number
                else:
                    raise FunctionalException(f"Enum type '{enum_type_descriptor.full_name}' has no name '{name_or_value}' (possible names: {[k for k in enum_type_descriptor.values_by_name]})")
            else:
                return name_or_value
        
        def __load_enum_data(self, enum_type_fullname, enum_type_descriptor):
            if enum_type_fullname in self.__enum_data_by_fullname:
                raise TechnicalException(f"Data of enum type '{enum_type_fullname}' was already loaded")
            data = NamedTuple('EnumData', values_by_number=dict)
            data.values_by_number = {}
            for value in enum_type_descriptor.values:
                if value.number not in data.values_by_number or value.index > data.values_by_number[value.number].index:
                    data.values_by_number[value.number] = value
            self.__enum_data_by_fullname[enum_type_fullname] = data
        
        def __set_object_value(self, field_obj, value, raise_exception=True):
            res = False
            for hpt in self.__registered_types:
                if hpt.is_instance_of(field_obj):
                    res = hpt.set_object_value(field_obj, value, raise_exception=raise_exception)
                    if res:
                        break
                    
            if not res and raise_exception:
                field_classes = inspect.getmro(type(field_obj))
                registered_types_str = ', '.join([f"{t} ({t.protobuf_class()})" for t in self.__registered_types])
                raise TechnicalException(f"Failed to manage type of field {Typing.get_object_class_fullname(field_obj)} (classes: {field_classes}) with registered types (and internal types): {registered_types_str}")
            return res
        
        def __set_object_repeated_field(self, obj, field_name, values, add_index=False):
            if not isinstance(values, list):
                raise FunctionalException(f"For repeated field '{field_name}', parameter 'values' is not a list (type: {Typing.get_object_class_fullname(values)})")
            
            for index, value in enumerate(values):
                self.__set_object_repeated_field_by_index(obj, field_name, index, value, add_index)
        
        def __set_object_repeated_field_by_index(self, obj, field_name, index, value, add_index=False):
            repeated_field_obj = getattr(obj, field_name)
            if not ProtobufMessages.is_object_repeated(repeated_field_obj):
                raise FunctionalException(f"Field '{field_name}' is not a repeated in type '{self.get_object_type_fullname(obj)}'")
            self.__check_field_value_to_set_is_valid(value)
            
            field_descr = self.__get_object_field_descriptor(obj, field_name)
            if field_descr.message_type is not None:
                # Get field object at given index
                if len(repeated_field_obj) < index + 1:
                    if add_index:
                        for _ in range(index + 1 - len(repeated_field_obj)):
                            obj_at_index = repeated_field_obj.add()
                    else:
                        raise FunctionalException(f"Index {index} exceeds repeated length {len(repeated_field_obj)} (field '{field_name}' in type '{self.get_object_type_fullname(obj)}')")
                else:
                    obj_at_index = repeated_field_obj[index]
                    
                # Set field object value
                try:
                    self.__set_object_value(obj_at_index, value)
                except Exception as exc:
                    raise TechnicalException(f"Unmanaged set of value [{value}] (type: {Typing.get_object_class_fullname(value)}) in repeated field of type '{self.get_object_type_fullname(obj_at_index)}' (at index {index} of field '{field_name}' in object of type '{self.get_object_type_fullname(obj)}' [{obj}])\n  -> {exc.message}") from exc
            else:
                if field_descr.enum_type is not None:
                    value_to_set = self.__get_enum_value(value, enum_type_descriptor=field_descr.enum_type)
                else:
                    value_to_set = value
                    
                if len(repeated_field_obj) < index:
                    if add_index:
                        for _ in range(index - len(repeated_field_obj)):
                            repeated_field_obj.append(type(value)())
                    else:
                        raise FunctionalException(f"Index {index} exceeds repeated length {len(repeated_field_obj)} (field '{field_name}' in object [{obj}])")
                    
                if len(repeated_field_obj) == index:
                    if add_index:
                        try:
                            repeated_field_obj.append(value_to_set)
                        except Exception as exc:
                            raise TechnicalException(f"Failed to add value [{value_to_set}] in repeated field at index {index} (field '{field_name}' in type '{self.get_object_type_fullname(obj)}')") from exc
                    else:
                        raise FunctionalException(f"Index {index} exceeds repeated length {len(repeated_field_obj)} (field '{field_name}' in type '{self.get_object_type_fullname(obj)}')")
                else:
                    try:
                        repeated_field_obj[index] = value_to_set
                    except Exception as exc:
                        raise TechnicalException(f"Failed to set value [{value_to_set}] in repeated field at index {index} (field '{field_name}' in type '{self.get_object_type_fullname(obj)}')") from exc
                
        def __split_attribute_fullname(self, attr_fullname):
            m = self.__regex_attribute_fullname.match(attr_fullname)
            res = [m.group(1), m.group(2)]
            # logger.trace(f"++++++++++ __split_attribute_fullname('{attr_fullname}') => {res}")
            return res
    
        def __represent_descriptor(self, descr, indent=0):
            res_str = [str(type(descr))]
            for name, value in Typing.get_object_attributes(descr):
                if "_by_" in name:
                    res_str.append(f"    {name}:")
                    for el in value:
                        res_str.append(f"        {el}: {self.__represent_descriptor(value[el], 8) if 'Descriptor' in str(type(value[el])) else value[el]}")
                else:
                    # if name == "containing_type":
                        res_str.append(f"    {name}: {value}")
                    # else:
                    #     res_str.append(f"    {name}: {self.__represent_descriptor(value, 4) if 'Descriptor' in str(type(value)) else value}")
            return Tools.indent_string(indent, "\n".join(res_str))
            
        def _register_types(self):
            # Duration must be registered before Message
            from holado_protobuf.ipc.protobuf.types.google.protobuf import Duration as hpt_Duration
            self.register_type(hpt_Duration)
    
            # Timestamp must be registered before Message
            from holado_protobuf.ipc.protobuf.types.google.protobuf import Timestamp as hpt_Timestamp
            self.register_type(hpt_Timestamp)
    
            from holado_protobuf.ipc.protobuf.types.google.protobuf import Message as hpt_Message
            self.register_type(hpt_Message)
    
        def register_type(self, protobuf_type, index=None):
            """
            Register a holado protobuf type (subclass of holado_protobuf.ipc.protobuf.abstracts.type.Type)
            """
            if protobuf_type in self.__registered_types:
                raise TechnicalException(f"Protobuf type '{protobuf_type.__class__.__name__}' is already registered")
            
            if index is not None:
                self.__registered_types.insert(index, protobuf_type)
            else:
                self.__registered_types.append(protobuf_type)





# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_grpc.api.rpc.grpc_client import GRpcClient
import logging
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_core.common.tools.tools import Tools
from holado_system.system.global_system import GlobalSystem
import queue
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


if GRpcClient.is_available():

    def __get_session_context():
        return SessionContext.instance()
    
    def __get_scenario_context():
        return __get_session_context().get_scenario_context()
    
    def __get_text_interpreter():
        return __get_scenario_context().get_text_interpreter()
    
    def __get_variable_manager():
        return __get_scenario_context().get_variable_manager()
    
    def __get_grpc_services():
        return __get_session_context().grpc_services

        
    @Given(r"(?P<var_name>{Variable}) = new gRPC client")
    def step_impl(context, var_name):
        var_name = StepTools.evaluate_variable_name(var_name)
        table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
        kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)
        
        client = __get_session_context().grpc_manager.new_client(None, **kwargs)
        
        __get_variable_manager().register_variable(var_name, client)
        
    @Given(r"(?P<var_name>{Variable}) = service names \(gRPC client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        client = StepTools.evaluate_variable_value(var_client)
        
        service_names = client.internal_client.service_names
        
        __get_variable_manager().register_variable(var_name, service_names)
        
    @Step(r"(?P<var_name>{Variable}) = input type fullname of request (?P<service_method>{Str}) \(gRPC client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, service_method, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        service_method = StepTools.evaluate_scenario_parameter(service_method)
        client = StepTools.evaluate_variable_value(var_client)
        
        service_name, method_name = service_method.rsplit('.', 1)
        res = client.get_request_data_type_fullname(service_name, method_name)
        
        __get_variable_manager().register_variable(var_name, res)
        
    @When(r"(?P<var_name>{Variable}) = request (?P<service_method>{Str}) \(gRPC client: (?P<var_client>{Variable})(?: ; with Protobuf (?P<with_proto_request>request)?(?: & )?(?P<with_proto_response>response)?)?\)")
    def step_impl(context, var_name, service_method, var_client, with_proto_request, with_proto_response):
        var_name = StepTools.evaluate_variable_name(var_name)
        service_method = StepTools.evaluate_scenario_parameter(service_method)
        client = StepTools.evaluate_variable_value(var_client)
        table = BehaveStepTools.get_step_table(context)
        
        service_name, method_name = service_method.rsplit('.', 1)
        request_data = client.build_request_data(service_name, method_name, params_table=table, as_proto=with_proto_request is not None)
        res = client.request(service_name, method_name, request_data, raw_output=with_proto_response is not None)
        
        __get_variable_manager().register_variable(var_name, res)
        
    @When(r"(?P<var_name>{Variable}) = request (?P<service_method>{Str}) with data (?P<request_data>{Str}) \(gRPC client: (?P<var_client>{Variable})(?: ; with Protobuf (?P<with_proto_response>response)?)?\)")
    def step_impl(context, var_name, service_method, request_data, var_client, with_proto_response):
        var_name = StepTools.evaluate_variable_name(var_name)
        service_method = StepTools.evaluate_scenario_parameter(service_method)
        request_data = StepTools.evaluate_scenario_parameter(request_data)
        client = StepTools.evaluate_variable_value(var_client)
        
        service_name, method_name = service_method.rsplit('.', 1)
        res = client.request(service_name, method_name, request_data, raw_output=with_proto_response is not None)
        
        __get_variable_manager().register_variable(var_name, res)
        
    @Step(r"(?P<var_name>{Variable}) = data for request (?P<service_method>{Str}) \(gRPC client: (?P<var_client>{Variable})(?P<as_proto> ; as Protobuf)?\)")
    def step_impl(context, var_name, service_method, var_client, as_proto):
        var_name = StepTools.evaluate_variable_name(var_name)
        service_method = StepTools.evaluate_scenario_parameter(service_method)
        client = StepTools.evaluate_variable_value(var_client)
        table = BehaveStepTools.get_step_table(context)
        
        service_name, method_name = service_method.rsplit('.', 1)
        res = client.build_request_data(service_name, method_name, params_table=table, as_proto=as_proto is not None)
        
        __get_variable_manager().register_variable(var_name, res)
        
    @When(r"request (?P<service_method>{Str}) for all of queue (?P<var_queue>{Variable}) \(gRPC client: (?P<var_client>{Variable})(?: ; rate log period: (?P<rate_period_s>{Int}) s)?\)")
    def step_impl(context, service_method, var_queue, var_client, rate_period_s):
        service_method = StepTools.evaluate_scenario_parameter(service_method)
        q = StepTools.evaluate_variable_value(var_queue)
        q_name = StepTools.evaluate_variable_name(var_queue)
        client = StepTools.evaluate_variable_value(var_client)
        client_name = StepTools.evaluate_variable_name(var_client)
        rate_period_s = StepTools.evaluate_scenario_parameter(rate_period_s)
        
        counter = 0
        beg = Tools.timer_s()
        c_beg, t_beg = counter, beg
        log_counter = 0
        service_name, method_name = service_method.rsplit('.', 1)
        
        try:
            while True:
                # Get next message and publish
                request_data = q.get()
                if q.is_sentinel(request_data):
                    break
                counter += 1
                
                try:
                    # Note: according internal library implementation, less actions are done with raw_output=True
                    client.request(service_name, method_name, request_data, raw_output=True)
                except Exception as exc:
                    logger.warning(f"[Queue {q_name} -> Client {client_name}] Failed to request data {counter}: [{Typing.get_object_class_fullname(exc)}] {str(exc)}")
                    # logger.warning(f"[Queue {q_name} -> Publisher {pub_name}] Failed to publish message {counter}: [{Typing.get_object_class_fullname(exc)}] {Tools.represent_exception(exc)}")
                finally:
                    q.task_done()
                
                # Log rate if needed
                if rate_period_s and counter % 10 == 0:
                    t_end = Tools.timer_s()
                    if t_end > t_beg + rate_period_s:
                        c_end, s_end = counter, q.qsize()
                        rate = (c_end - c_beg) / (t_end - t_beg)
                        log_counter += 1
                        logger.print(f"[Queue {q_name} -> Client {client_name}] Rate: {int(rate)} msg/s ; Nb messages: {counter} ; Queue size: {s_end}")
                        c_beg, t_beg = c_end, t_end
                        
                # Yield processor in case of multithreading
                if counter % 10 == 0:
                    GlobalSystem.yield_processor()
        except queue.Empty:
            # Without block, or with timeout, this exception occurs when queue is empty
            pass
        
        if rate_period_s:
            end = Tools.timer_s()
            rate = counter / (end - beg)
            logger.print(f"[Queue {q_name} -> Client {client_name}] Mean rate: {int(rate)} msg/s ; Nb messages: {counter}")




    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.technical_exception import TechnicalException
import logging
from holado_core.common.finders.tools.find_context import ContainerFindContext
from holado_core.common.finders.tools.find_parameters import FindParameters
from holado_core.common.finders.tools.enums import FindType
from holado_core.common.finders.tree_finder import TreeFinder
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


class AfterInTreeFinder(TreeFinder):
    """ Finder that searches through tree an element after given one.
    It goes through parents, and for each parent, look in children after the child coming from if element is found.
    The element is searched using given finder from each possible child.
    """
    
    def __init__(self, finder_element, description=None):
        super().__init__(description)
        self.find_type = FindType.Custom
        
        self.__finder_element = finder_element
    
    def _find_all_container(self, find_context:ContainerFindContext, find_parameters:FindParameters):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"FinderHtmlAfterInTree[{self.element_description}]._find_all_container begin  It should be replaced by another find method for performance reasons")

        res = []
        sub_find_parameters = find_parameters.get_next_level(False)
        
        current_context = find_context
        while len(res) == 0:
            # Find parent
            parent = self._find_parent(current_context, sub_find_parameters)
            if parent is None:
                # Root parent is reached, stop searching
                break
            
            # Get children
            children = self._find_children(current_context, sub_find_parameters, parent)
            
            # Get index in children
            index_in_children = children.index(current_context.container)
            if index_in_children < 0:
                raise TechnicalException("Unable to find element in children of parent")
            
            # Search through each possible child
            for index in range(index_in_children + 1, len(children)):
                child = children[index]
                
                # Check child
                try:
                    if self.__finder_element.is_valid_input(child, current_context, sub_find_parameters):
                        self._add_candidate(res, child)
                        continue
                except NotImplementedError:
                    # check this child is not relevant, continue code
                    pass
                
                # Check in child
                res_child = self.__finder_element.find_all_in(child, find_context=current_context, find_parameters=sub_find_parameters)
                if len(res_child) > 0:
                    self._add_candidates(res, res_child)
                    continue
            
            # Prepare next loop
            current_context = current_context.withContainer(parent)
        
        return res
    
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.technical_exception import TechnicalException
import logging
from holado_core.common.tools.tools import Tools
import os
from holado_core.tools.abstracts.blocking_command_service import BlockingCommandService
from holado_system.system.command.command import CommandStates
from holado_core.common.handlers.wait import WaitFuncResultVerifying
from holado.common.handlers.object import DeleteableObject

logger = logging.getLogger(__name__)

try:
    import django  # @UnusedImport
    with_django = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"DjangoServer is not available. Initialization failed on error: {exc}")
    with_django = False


class DjangoServer(DeleteableObject):
    """
    Django server
    """
    
    @classmethod
    def is_available(cls):
        return with_django
    
    def __init__(self, name, django_project_path, port=8000, runserver_args=None):
        super().__init__(name)
        
        self.__django_project_path = django_project_path
        self.__runserver_port = port
        self.__runserver_args = runserver_args
        self.__project_command = None
        
        if not os.path.exists(django_project_path):
            raise TechnicalException(f"Django project doesn't exist (project path: '{django_project_path}')")
    
    def _delete_object(self):
        if self.__project_command is not None and self.__project_command.status == CommandStates.Running:
            self.stop()

    @property
    def django_project_path(self):
        return self.__django_project_path

    @property
    def runserver_port(self):
        return self.__runserver_port

    @property
    def runserver_args(self):
        return self.__runserver_args
    
    def start(self):
        self.__start_server_by_command()
    
    def __start_server_by_command(self):
        self.__project_command = self._new_project_command()
        self.__project_command.start()
        self._wait_until_server_is_reachable()
    
    def _new_project_command(self):
        manage_path = os.path.join(self.django_project_path, "manage.py")
        cmd = f"python {manage_path} runserver {self.runserver_port}"
        if self.runserver_args:
            cmd += f" {self.runserver_args}"
        
        res = BlockingCommandService(f"Command running Django server '{self.name}'", cmd)
        res.auto_stop = True
        
        return res
        
    def _wait_until_server_is_reachable(self):
        import requests
        url = f"http://127.0.0.1:{self.runserver_port}"
        redo = WaitFuncResultVerifying("server is reachable", 
                                       lambda: requests.get(url), 
                                       lambda result: result and result.status_code == 200 )
        redo.ignoring(Exception)
        redo.polling_every(0.01)
        redo.execute()
        
    def stop(self):
        if self.__project_command is None or self.__project_command.status != CommandStates.Running:
            raise TechnicalException(f"Django server of project '{self.name}' is not running (status: {self.__project_command.status if self.__project_command else 'Unkown'})")
        self.__project_command.stop()
    
    def join(self, timeout=None):
        self.__project_command.join(timeout)
    







from django.shortcuts import render
from django.contrib.auth.models import User, Group
from rest_framework import viewsets
from rest_framework import permissions
from api_rest.api1.serializers import UserSerializer, GroupSerializer


class UserViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows users to be viewed or edited.
    """
    queryset = User.objects.all().order_by('-date_joined')
    serializer_class = UserSerializer
    permission_classes = [permissions.IsAuthenticated]


class GroupViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows groups to be viewed or edited.
    """
    queryset = Group.objects.all()
    serializer_class = GroupSerializer
    permission_classes = [permissions.IsAuthenticated]
    



# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

# This file is a helper to build scripts using HolAdo solution, with CSV files as input.
# Usually, it is copied and adapted for specific usage.


import logging
from holado_python.common.tools.datetime import DateTime
from holado_helper.script.action_script import ActionScript, ActionScriptActor, ActionScriptFile, ActionRegistry
from holado_helper.script.action import BehaveActionInfo
from holado_core.common.actors.actions import Action
import copy

logger = logging.getLogger(__name__)


class BehaveActionScriptFile(ActionScriptFile):
    def __init__(self, file_ext):
        super().__init__(file_ext)


class BehaveActionRegistry(ActionRegistry):
    def __init__(self, name):
        super().__init__(name)
    
    def register_action(self, action, parameters_default_values=None, parameters_replacements=None, behave_tags=None, **kwargs):
        action_kwargs = {}
        if behave_tags is not None:
            action_kwargs['behave_tags'] = behave_tags
        action_kwargs.update(kwargs)
        super().register_action(action, parameters_default_values=parameters_default_values, parameters_replacements=parameters_replacements, **action_kwargs)
    
    def get_action_behave_tags(self, action, action_parameters=None):
        return self.get_action_kwargs(action).get('behave_tags', None)


class BehaveActionScriptActor(ActionScriptActor):
    def __init__(self, module_name, action_registry=None):
        super().__init__(module_name, action_registry)
    
    def act_process_action(self, action_info, action_script):
        def process_action():
            from holado.common.context.session_context import SessionContext
            from holado_helper.script.action import BehaveActionRunner
            
            var_manager = SessionContext.instance().multitask_manager.get_thread_context().get_variable_manager()
            behave_runner = BehaveActionRunner(action_info, var_manager)
            res_main = behave_runner.run()
            
            res = copy.deepcopy(action_info)
            if res.params is None:
                res.params = {}
            res.params['STATUS'] = 'SUCCESS' if res_main else 'ERROR'
            
            return res
        return Action(action_info.action, process_action)


class BehaveActionScript(ActionScript):
    def __init__(self, work_dir_path=None, with_input=True, with_output=True, with_template=True, with_action_parameter=True, 
                 actor=None, action_registry=None, script_file=None):
        if actor is None:
            actor = BehaveActionScriptActor("default_behave", action_registry)
        super().__init__(work_dir_path=work_dir_path, with_input=with_input, with_output=with_output, with_template=with_template, 
                         with_action_parameter=with_action_parameter, actor=actor, action_registry=action_registry, script_file=script_file)
    
    def build_action_tags(self, action, action_args=None):
        return list(self.action_registry.get_action_behave_tags(action))
    
    def _build_behave_args(self, tags=None):
        res = f'--no-source --no-skipped --format null --no-summary --no-capture --no-capture-stderr --no-logcapture'
        if tags:
            for tag in tags:
                res += f' -t {tag}'
        return res
    
    def new_action_info_from_args(self):
        action = self.args.action
        static_params = self.action_registry.get_action_static_parameters_values(action)
        behave_args = self._build_behave_args(self.action_registry.get_action_behave_tags(action))
        return self.new_action_info(action, static_params=static_params, behave_args=behave_args)
    
    def new_action_info(self, action, uid=None, behave_args=None, params=None, static_params=None, index=0):
        return BehaveActionInfo(uid=uid, action=action, params=params, static_params=static_params, index=index, behave_args=behave_args)
    





# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

# This file is a helper to build scripts using HolAdo solution, with CSV files as input.
# Usually, it is copied and adapted for specific usage.


import logging
import os
import re
from holado_helper.script.script import Script
from holado_core.common.exceptions.technical_exception import TechnicalException

logger = logging.getLogger(__name__)


class InputOutputScript(Script):
    
    def __init__(self, work_dir_path=None, with_input=True, with_output=True, with_template=False):
        super().__init__()
        self.__work_dir_path = work_dir_path
        self.__with_input = with_input
        self.__with_output = with_output
        self.__with_template = with_template
    
    @property
    def with_input(self):
        return self.__with_input
    
    @property
    def with_output(self):
        return self.__with_output
    
    @property
    def with_template(self):
        return self.__with_template
    
    def initialize(self):
        super().initialize()
        
        if self.__work_dir_path is not None:
            self.change_working_dir(self.__work_dir_path)
    
    def is_matching_file_path(self, input_file_path, raise_exception=False):
        raise NotImplementedError
    
    def is_matching_input_file_path(self, input_file_path, raise_exception=False):
        return self.is_matching_file_path(input_file_path, raise_exception)
    
    def is_matching_output_file_path(self, input_file_path, raise_exception=False):
        return self.is_matching_file_path(input_file_path, raise_exception)
    
    def _new_argument_parser(self, description=None, additional_arguments=None, add_help=True):
        parser = super()._new_argument_parser(description=description, additional_arguments=additional_arguments, add_help=add_help)
        
        if self.__with_input:
            parser.add_argument('-i', '--input', dest='input_filename', default=None,
                           help='Input file name')
        
        if self.__with_output:
            parser.add_argument('-o', '--output', dest='output_filename', default=None,
                           help='Output file name (no output by default)')
        
        if self.__with_template:
            parser.add_argument('-ot', '--output-template', dest='output_template', default=False, action="store_true",
                           help='If specified, output a template')
        
        return parser
    
    def _build_input_path(self, input_filename, cwd):
        if input_filename is None:
            logger.print(f"Input filename must be specified with parameter '-i' (-h for help)")
            exit(1)
            
        input_path = self._build_path(input_filename, cwd)
        if not os.path.exists(input_path):
            logger.print(f"Input file '{input_path}' doesn't exist")
            exit(1)
        
        return input_path
    
    def _build_output_template_path(self):
        """Return output template path.
        It should be overriden, and call _build_output_path with appropriate base_default_path argument.
        """
        return self._build_output_path(self.args.output_filename, None, self.user_working_dir)
    
    def _build_output_path(self, output_filename, base_default_path, cwd, raise_exception=True):
        res = None
        if output_filename is not None:
            res = self._build_path(output_filename, cwd)
        elif base_default_path is not None:
            output_path_base, with_suffix, ext = self._build_output_path_base(base_default_path)
            if with_suffix:
                res = self._build_path(f"{output_path_base}-SUCCESS{ext}", cwd)
            else:
                res = self._build_path(f"{output_path_base}{ext}", cwd)
        elif raise_exception:
            raise TechnicalException("Failed to build output path, one of parameters 'output_filename' and 'base_default_path' must be defined")
        return res
    
    def _build_output_path_base(self, output_path):
        res, ext = os.path.splitext(output_path)
        with_suffix = False
        if res.endswith('-SUCCESS'):
            res = res[:-len('-SUCCESS')]
            with_suffix = True
        else:
            m = re.match(r'^(.*)-FAIL(_\d+)?$', res)
            if m:
                res = m.group(1)
                with_suffix = True
        return res, with_suffix, ext
    
    def _build_output_path_from_base(self, output_path_base, ext):
        output_path = f"{output_path_base}{ext}"
        if os.path.exists(output_path):
            suffix_index = 1
            output_path = f"{output_path_base}_{suffix_index}{ext}"
            while os.path.exists(output_path):
                suffix_index += 1
                output_path = f"{output_path_base}_{suffix_index}{ext}"
        return output_path
    





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import ssl
from holado_core.common.tools.tools import Tools
from holado_json.ipc.json import set_object_attributes_with_json_dict
from holado_core.common.exceptions.technical_exception import TechnicalException
import os
import copy
from ssl import Purpose
from holado_system.system.command.command import Command, CommandStates
from holado.common.context.session_context import SessionContext
import json

logger = logging.getLogger(__name__)


class SslManager(object):
    """
    Helper for ssl module.
    
    It is implemented internally with standard library ssl.
    """
    
    @classmethod
    def __get_path_manager(cls):
        return SessionContext.instance().path_manager
    
    @classmethod
    def new_ssl_context(cls, server_side=False, ssl_kwargs=None, context_kwargs=None):
        res = None
        
        if ssl_kwargs is None:
            ssl_kwargs = {}
        kwargs = copy.copy(ssl_kwargs)
        
        try:
            activate_ssl = kwargs.pop('activate', True)
            if activate_ssl:
                if Tools.has_sub_kwargs(kwargs, 'create_default_context.'):
                    create_default_context_kwargs = Tools.pop_sub_kwargs(kwargs, 'create_default_context.')
                else:
                    create_default_context_kwargs = {}
                
                purpose = Purpose.CLIENT_AUTH if server_side else Purpose.SERVER_AUTH
                res = ssl.create_default_context(purpose=purpose, **create_default_context_kwargs)
                
                # res.set_default_verify_paths()
                
                if context_kwargs:
                    c_kwargs = copy.copy(context_kwargs)
                    
                    ciphers = c_kwargs.pop('ciphers', None)
                    if ciphers is not None:
                        if ciphers == 'OPENSSL_CIPHERS':
                            ciphers = SslManager.get_openssl_ciphers()
                            logger.debug(f"Using openssl ciphers: {ciphers}")
                        res.set_ciphers(ciphers)
                    if Tools.has_sub_kwargs(c_kwargs, 'load_cert_chain.'):
                        load_cert_chain_kwargs = Tools.pop_sub_kwargs(c_kwargs, 'load_cert_chain.')
                        res.load_cert_chain(**load_cert_chain_kwargs)
                    if Tools.has_sub_kwargs(c_kwargs, 'load_verify_locations.'):
                        load_verify_locations_kwargs = Tools.pop_sub_kwargs(c_kwargs, 'load_verify_locations.')
                        res.load_verify_locations(**load_verify_locations_kwargs)
                    
                    # Set context attributes with remaining c_kwargs
                    if len(c_kwargs) > 0:
                        set_object_attributes_with_json_dict(res, c_kwargs)
                
                logger.debug(f"Default verify paths: {ssl.get_default_verify_paths()})")
                dvp = ssl.get_default_verify_paths()
                if res.verify_mode != ssl.CERT_NONE and dvp.cafile is None and 'create_default_context.cafile' not in ssl_kwargs and 'load_verify_locations.cafile' not in context_kwargs:
                    msg_list = [f"No CA file is defined, it is not possible to verify certificates.",
                                f"Most common solutions:",
                                f"  - Configure to not verify certificates, by adding/setting in table: | 'ssl.context.verify_mode' | ssl.CERT_NONE |",
                                f"  - Be sure OpenSSL default CA file exists on system (configured path: '{dvp.openssl_cafile}'). Example of command (on Ubuntu): sudo ln -s /etc/ssl/certs/ca-certificates.crt {dvp.openssl_cafile}",
                                f"  - Configure a CA file at context creation, by adding/setting in table: | 'ssl.create_default_context.cafile' | <path to CA file> |",
                                f"  - Configure a CA file after context creation, by adding/setting in table: | 'ssl.context.load_verify_locations.cafile' | <path to CA file> |",
                                f"Note: step 'Given CACERTS_PATH = CA certs file path (from certifi package)' can be used to get the path to CA file coming with 'certifi' package that is supposed to be installed when using ssl module",
                                ]
                    raise TechnicalException("\n".join(msg_list))
                try:
                    logger.debug(f"Loaded CA certificates: {res.get_ca_certs()})")
                except NotImplementedError:
                    # Depending on environment, get_ca_certs can be not implemented (ex: python 3.10 and pip v25.1.1)
                    pass
                logger.debug(f"Loaded ciphers: {res.get_ciphers()})")
                
        except Exception as exc:
            msg = f"Failed to create SSL context with parameters ({ssl_kwargs}, {context_kwargs}): {exc}"
            logger.error(msg)
            raise TechnicalException(msg) from exc
        
        # Verify all kwargs were applied
        if len(kwargs) > 0:
            raise TechnicalException(f"Unmanaged ssl parameters: {kwargs}")
        
        return res
    
    @classmethod
    def get_openssl_ciphers(cls):
        import subprocess
        output = subprocess.run(["openssl", "ciphers"], capture_output=True).stdout
        output_str = output.decode("utf-8")
        return output_str.strip().split("\n")[0]
    
    @classmethod
    def get_default_certificates(cls):
        dvp = ssl.get_default_verify_paths()
        if not os.path.exists(dvp.cafile):
            raise TechnicalException(f"openssl CA file '{dvp.cafile}' doesn't exist. Example of resolution command (on Ubuntu): sudo ln -s /etc/ssl/certs/ca-certificates.crt {dvp.cafile}")
        return (dvp.cafile, dvp.capath)
    
    @classmethod
    def get_certifi_ca_certs_file_path(cls):
        import certifi
        return certifi.where()
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.handlers.element_holder import ElementHolder
from holado_core.common.exceptions.technical_exception import TechnicalException
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.alert import Alert
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.common.by import By


class SeleniumBy(object):
    """ Information on a Selenium By.
    """
    
    def __init__(self, by:By, value=None):
        """
        @param by Selenium locator strategy
        @param value Selenium locator value
        """
        self.__by = by
        self.__value = value

    @property
    def by(self):
        """
        @return Selenium locator strategy
        """
        return self.__by
    
    @property
    def value(self):
        """
        @return Selenium locator value
        """
        return self.__value
    
    def __repr__(self)->str:
        return f"{self.by.replace(' ', '_')}({self.value})"
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import socket
from _socket import SHUT_RDWR, SOCK_STREAM, error
from holado.common.handlers.object import DeleteableObject
import abc
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.ssl.ssl import SslManager
import threading
import types
from abc import abstractmethod
from holado.common.handlers.undefined import undefined_argument, default_value
import copy
from holado_python.standard_library.typing import Typing
import ssl
import select
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.handlers.wait import WaitFuncResultVerifying

logger = logging.getLogger(__name__)


class SocketConfig(object):
    default_idle_sleep_delay = 0.01
    default_read_bufsize = 1024


class Socket(DeleteableObject):
    """
    Base class for socket objects.
    
    It is implemented internally with standard library socket.
    """
    __metaclass__ = abc.ABCMeta
    
    @classmethod
    def create(cls, name, internal_socket):
        res = cls.__new__(cls, name=name)
        res.__init__(name=name)
        res._set_internal_socket(internal_socket)
        return res
    
    @classmethod
    def create_connection(cls, address, timeout=undefined_argument,
                          source_address=None, do_connect=True):
        """Connect to *address* and return the socket object.
    
        Convenience function.  Connect to *address* (a 2-tuple ``(host,
        port)``) and return the socket object.  Passing the optional
        *timeout* parameter will set the timeout on the socket instance
        before attempting to connect.  If no *timeout* is supplied, the
        global default timeout setting returned by :func:`getdefaulttimeout`
        is used.  If *source_address* is set it must be a tuple of (host, port)
        for the socket to bind as a source address before making the connection.
        A host of '' or port 0 tells the OS to use the default.
        
        Note: Implementation is a copy of socket.create_connection, but allowing to not connect.
              If do_connect==True, implementation is identical to socket.create_connection.
        """
    
        host, port = address
        err = None
        for res in socket.getaddrinfo(host, port, 0, SOCK_STREAM):
            af, socktype, proto, canonname, sa = res  # @UnusedVariable
            sock = None
            try:
                sock = socket.socket(af, socktype, proto)
                if timeout is not undefined_argument:
                    sock.settimeout(timeout)
                if source_address:
                    sock.bind(source_address)
                if do_connect:
                    sock.connect(sa)
                # Break explicitly a reference cycle
                err = None
                return sock
    
            except error as _:
                err = _
                if sock is not None:
                    sock.close()
    
        if err is not None:
            try:
                raise err
            finally:
                # Break explicitly a reference cycle
                err = None
        else:
            raise error("getaddrinfo returns an empty list")
    
    def __init__(self, *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument):
        """Socket constructor
        @param name: Socket name
        @param create_ipv4_socket_kwargs: arguments to create an IPv4 socket
        @param idle_sleep_delay: delay to sleep when socket is idle (used in some clients/servers to control CPU resource impact ; default: 0.01 s)
        """
        if name is None and create_ipv4_socket_kwargs is not None:
            if not set(create_ipv4_socket_kwargs.keys()).issuperset({'host', 'port'}):
                raise FunctionalException(f"Parameters 'host' and 'port' must be defined")
            name = f"{Typing.get_object_class_name(self)}({create_ipv4_socket_kwargs['host']}:{create_ipv4_socket_kwargs['port']})"
        
        super().__init__(name)
        self.__socket = None
        self.__is_started = False
        self.__start_thread = None
        self.__stop_on_close = None
        self.__idle_sleep_delay = idle_sleep_delay if idle_sleep_delay is not undefined_argument else SocketConfig.default_idle_sleep_delay
        
        # SSL management
        self.__ssl_kwargs_tuple = None
        self.__ssl_context = None
        self.__is_ssl_handshake_done = False
        
        if create_ipv4_socket_kwargs is not None:
            self.create_ipv4_socket(**create_ipv4_socket_kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def _delete_object(self):
        self.close()
    
    @property    
    def internal_socket(self) -> socket.socket:
        return self.__socket
    
    @property
    def is_started(self):
        return self.__is_started
    
    @property
    def is_with_ssl(self):
        return self.__ssl_context is not None
    
    @property
    def _ssl_kwargs(self):
        return copy.copy(self.__ssl_kwargs_tuple[0]) if self.__ssl_kwargs_tuple and self.__ssl_kwargs_tuple[0] is not None else None
    
    @property
    def _ssl_context_kwargs(self):
        return copy.copy(self.__ssl_kwargs_tuple[1]) if self.__ssl_kwargs_tuple and self.__ssl_kwargs_tuple[1] is not None else None
    
    @property
    def _ssl_wrap_socket_kwargs(self):
        return copy.copy(self.__ssl_kwargs_tuple[2]) if self.__ssl_kwargs_tuple and self.__ssl_kwargs_tuple[2] is not None else None
    
    @property
    def _ssl_context(self):
        return self.__ssl_context
    
    @property
    def _idle_sleep_delay(self):
        return self.__idle_sleep_delay
    
    def _set_internal_socket(self, socket, is_ssl_handshake_done_on_connect=True):
        self.__socket = socket
        self.__is_ssl_handshake_done = is_ssl_handshake_done_on_connect
    
    @abc.abstractmethod
    def create_ipv4_socket(self, host, port, **kwargs):
        raise NotImplementedError()
    
    def close(self):
        # Stop socket if needed
        if self.is_started and self.__stop_on_close:
            self.stop()
        
        # Close socket
        if self.__socket is not None:
            try:
                self.__socket.shutdown(SHUT_RDWR)
            except OSError as exc:
                if 'Errno 9' in str(exc):
                    logger.info(f"Got following error on socket shutdown (known to appear when connection is not open correctly or already closed): {exc}")
                elif 'Errno 107' in str(exc):
                    logger.info(f"Got following error on socket shutdown (known to appear sometimes during shutdown): {exc}")
                else:
                    raise exc
            finally:
                self.__socket.close()
                self.__socket = None
    
    @abstractmethod
    def start(self, *, read_bufsize=SocketConfig.default_read_bufsize, read_kwargs=None, write_kwargs=None):
        raise NotImplementedError()
    
    def _start_thread(self, thread, stop_on_close=True):
        self.__start_thread = thread
        self.__stop_on_close = stop_on_close
        
        # Start thread
        self.__start_thread.start()
        self.__is_started = True
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Started")
    
    def stop(self):
        if self.__start_thread is not None:
            if self.__start_thread.is_interruptable:
                self.__start_thread.interrupt()
            self.__start_thread.join()
            self.__start_thread = None
            self.__is_started = False
            
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"[{self.name}] Stopped")
    
    def read(self, bufsize=SocketConfig.default_read_bufsize, **kwargs):
        flags = kwargs.get('flags', 0)
        res = self.internal_socket.recv(bufsize, flags)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"[{self.name}] Received [{res}] (type: {type(res)})")
        return res
    
    def write(self, data_bytes, loop_until_all_is_sent=True, **kwargs):
        if loop_until_all_is_sent:
            # Note: no return is done, since sendall return None or raise an exception on error
            self.internal_socket.sendall(data_bytes, **kwargs)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"[{self.name}] Sent all [{data_bytes}] (type: {type(data_bytes)})")
        else:
            res = self.internal_socket.send(data_bytes, **kwargs)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"[{self.name}] Sent {res}/{len(data_bytes)} bytes of [{data_bytes}] (type: {type(data_bytes)})")
            return res
    
    def _extract_ssl_kwargs(self, socket_kwargs):
        """Pop from socket kwargs: global ssl kwargs, SSLContext kwargs, and wrap_socket kwargs.
        """
        ssl_kwargs, context_kwargs, wrap_socket_kwargs = None, {}, {}
        if Tools.has_sub_kwargs(socket_kwargs, 'ssl.'):
            ssl_kwargs = Tools.pop_sub_kwargs(socket_kwargs, 'ssl.')
            if Tools.has_sub_kwargs(ssl_kwargs, 'context.'):
                context_kwargs = Tools.pop_sub_kwargs(ssl_kwargs, 'context.')
            if Tools.has_sub_kwargs(ssl_kwargs, 'wrap_socket.'):
                wrap_socket_kwargs = Tools.pop_sub_kwargs(ssl_kwargs, 'wrap_socket.')
        self.__ssl_kwargs_tuple = (ssl_kwargs, context_kwargs, wrap_socket_kwargs)
        return socket_kwargs
    
    def _new_ssl_context_if_required(self, server_side=False, **kwargs):
        """Return the remaining socket kwargs after ssl kwargs extraction.
        """
        res = self._extract_ssl_kwargs(kwargs)
        
        if self._ssl_kwargs is not None:
            self.__ssl_context = SslManager.new_ssl_context(server_side=server_side, ssl_kwargs=self._ssl_kwargs, context_kwargs=self._ssl_context_kwargs)
            
        return res
    
    def do_ssl_handshake(self):
        if self.is_with_ssl:
            if self.internal_socket.getblocking():
                self.internal_socket.do_handshake()
                # self.internal_socket.do_handshake(block=True)
            else:
                while True:
                    try:
                        self.internal_socket.do_handshake()
                        break
                    except ssl.SSLWantReadError:
                        select.select([self.internal_socket], [], [])
                    except ssl.SSLWantWriteError:
                        select.select([], [self.internal_socket], [])


    
    def ensure_ssl_handshake_is_done(self):
        if self.is_with_ssl and not self.__is_ssl_handshake_done:
            self.do_ssl_handshake()
            self.__is_ssl_handshake_done = True


class SocketClient(Socket):
    """
    Base class for socket clients.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument, do_run_with_recv=True, do_run_with_send=True):
        """Socket client constructor
        @param name: Socket client name
        @param create_ipv4_socket_kwargs: Arguments to create an IPv4 socket
        @param idle_sleep_delay: delay to sleep when socket is idle (used only in some clients to control CPU resource impact ; default: 0.01 s)
        @param do_run_with_recv: Define if recv is done in run process when client is started
        @param do_run_with_send: Define if send is done in run process when client is started
        """
        # Data used in start processing
        self.__data_lock = threading.Lock()
        self.__data = types.SimpleNamespace(
            in_bytes = b'',
            out_bytes = b'',
        )
        self.__with_recv = do_run_with_recv
        self.__with_send = do_run_with_send
        
        # Note: super().__init__ must be done after self.__data & others initialization since it can execute create_ipv4_socket_kwargs method
        super().__init__(name=name, create_ipv4_socket_kwargs=create_ipv4_socket_kwargs, idle_sleep_delay=idle_sleep_delay)
    
    @property
    def is_run_with_recv(self):
        return self.__with_recv
    
    @property
    def is_run_with_send(self):
        return self.__with_send
    
    @property
    def _data_lock(self):
        return self.__data_lock
    
    @property
    def _data(self):
        return self.__data
    
    def _start_thread(self, thread, stop_on_close=True):
        # Ensure SSL handshake is done before starting to receive and send data
        self.ensure_ssl_handshake_is_done()
        
        # Start thread
        super()._start_thread(thread, stop_on_close=stop_on_close)
    
    @property
    def received_data(self):
        with self.__data_lock:
            res = copy.copy(self.__data.in_bytes)
        return res
    
    @property
    def received_data_size(self):
        with self.__data_lock:
            res = len(self.__data.in_bytes)
        return res
    
    def reset_received_data(self):
        with self.__data_lock:
            self.__data.in_bytes = b''
    
    def read(self, bufsize=SocketConfig.default_read_bufsize, **kwargs):
        if self.is_started and self.is_run_with_recv:
            with self.__data_lock:
                if self.__data.in_bytes:
                    res = self.__data.in_bytes[:bufsize]
                    self.__data.in_bytes = self.__data.in_bytes[bufsize:]
                else:
                    res = b''
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"[{self.name}] Read data from received data: [{res}] (type: {type(res)} ; remaining data: {len(self.__data.in_bytes)})")
            return res
        else:
            return super().read(bufsize=bufsize, **kwargs)
    
    def write(self, data_bytes, loop_until_all_is_sent=True, **kwargs):
        if self.is_started and self.is_run_with_send:
            with self.__data_lock:
                self.__data.out_bytes += data_bytes
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"[{self.name}] Added data in data to send (total: {len(self.__data.out_bytes)})")
        else:
            return super().write(data_bytes, loop_until_all_is_sent=loop_until_all_is_sent, **kwargs)
        
    def wait_data(self, min_length=1, timeout_sec=default_value, do_raise_on_timeout=True):
        wait = WaitFuncResultVerifying("wait data", lambda: self.received_data_size, lambda r: r >= min_length)
        if timeout_sec is not default_value:
            wait.with_timeout(timeout_sec)
        if not do_raise_on_timeout:
            wait.without_raise_on_timeout()
        wait.execute()


class SocketServer(Socket):
    """
    Base class for socket servers.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, *, name=None, create_ipv4_socket_kwargs=None, idle_sleep_delay=undefined_argument):
        super().__init__(name=name, create_ipv4_socket_kwargs=create_ipv4_socket_kwargs, idle_sleep_delay=idle_sleep_delay)
    
    def accept(self):
        if self.internal_socket:
            conn, addr = self.internal_socket.accept()
            res = Socket.create(name=f"[{self.name}] Connection with {addr}", internal_socket=conn)
            return res, addr
        else:
            # Server is stopped
            raise TechnicalException(f"[{self.name}] Server socket is closed")
    
    @abc.abstractmethod
    def _process_received_data(self, data):
        raise NotImplementedError()







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2023 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_db.tools.db.clients.base.db_client import DBClient
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.tools import Tools
from holado_db.tools.db.clients.sqlite.sqlite_audit import SQLite3TriggerAuditManager

logger = logging.getLogger(__name__)

try:
    import sqlite3
    with_sqlite3 = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"SQLite3Client is not available. Initialization failed on error: {exc}")
    with_sqlite3 = False


class SQLite3Client(DBClient):
    @classmethod
    def is_available(cls):
        return with_sqlite3

    def __init__(self, name, connect_kwargs):
        super().__init__(name if name else 'SQLite3', connect_kwargs)
        
    def _new_audit_manager(self):
        return SQLite3TriggerAuditManager("Audit with triggers", self)
    
    def _get_base_exception_type(self):
        return sqlite3.Error
        
    def _connect(self, **kwargs):
        try:
            connect_kwargs = dict(kwargs)
            database = connect_kwargs.pop("database")
            return sqlite3.connect(database, **connect_kwargs)
        except Exception as exc:
            raise FunctionalException(f"Failed to connect with kwargs={kwargs}.\nUsage:\n{Tools.indent_string(4, sqlite3.connect.__doc__)}") from exc  # @UndefinedVariable
        
    def _get_sql_placeholder(self):
        return "?"
        
    def exist_table(self, table_name):
        result = self.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}'")
        return result is not None and result.nb_rows > 0
        
    def exist_table_index(self, table_name, index_name):
        result = self.execute(f"SELECT name FROM sqlite_master WHERE type='index' AND tbl_name='{table_name}' AND name='{index_name}'")
        return result is not None and result.nb_rows > 0

    def get_table_index_names(self, table_name):
        result = self.execute(f"SELECT name FROM sqlite_master WHERE type='index' AND tbl_name='{table_name}'", result_as_dict_list=True)
        return [r['name'] for r in result]




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.common.context.session_context import SessionContext
from holado.common.handlers.undefined import default

logger = logging.getLogger(__name__)


class CryptoManager(object):
    """
    Helper for cryptography management.
    """
    
    @classmethod
    def __get_path_manager(cls):
        return SessionContext.instance().path_manager
    
    @classmethod
    def encrypt_data_with_key(cls, data, key, data_padding_char=b'\x00', padding=default):
        """ Encrypt data with given asymmetric public key.
        Note: default padding is OAEP with sha256. If it doesn't match the key, padding parameter must be specified.
        """
        return cls._crypt_data_with_key(data, key.key_size, key.encrypt, data_padding_char, padding)
    
    @classmethod
    def decrypt_data_with_key(cls, data, key, data_padding_char=b'\x00', padding=default):
        """ Decrypt data with given asymmetric private key.
        Note: default padding is OAEP with sha256. If it doesn't match the key, padding parameter must be specified.
        """
        return cls._crypt_data_with_key(data, key.key_size, key.decrypt, data_padding_char, padding)

    @classmethod
    def _crypt_data_with_key(cls, data, key_size, key_method, data_padding_char=b'\x00', padding=default):
        """ Decrypt data with given asymmetric private key.
        Note: default padding is OAEP with sha256. If it doesn't match the key, padding parameter must be specified.
        """
        if padding is default:
            from cryptography.hazmat.primitives.asymmetric import padding
            from cryptography.hazmat.primitives import hashes
            padding = padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA256()), algorithm=hashes.SHA256(), label=None)
        
        # Decrypt data
        res = b''
        key_nb_bytes = key_size // 8
        for i in range(0, len(data), key_nb_bytes):
            cur_data = data[i*key_nb_bytes:(i+1)*key_nb_bytes]
            cur_len = len(cur_data)
            if cur_len < key_nb_bytes:
                cur_data += data_padding_char * (key_nb_bytes - cur_len)
            cur_res = key_method(cur_data, padding=padding)
            if cur_len < key_nb_bytes:
                cur_res = cur_res[:cur_len]
            res += cur_res
        
        return res

    @classmethod
    def cipher_data(cls, data, cipher=None, **cipher_kwargs):
        """ Cipher data with given symmetric cipher.
        """
        if cipher is None:
            cipher = cls.new_cipher(**cipher_kwargs)
        encryptor = cipher.encryptor()
        res = encryptor.update(data) + encryptor.finalize()
        return res

    @classmethod
    def decipher_data(cls, data, cipher=None, **cipher_kwargs):
        """ Decipher data with given symmetric cipher.
        """
        if cipher is None:
            cipher = cls.new_cipher(**cipher_kwargs)
        decryptor = cipher.decryptor()
        res = decryptor.update(data) + decryptor.finalize()
        return res
        

    @classmethod
    def new_cipher(cls, algorithm=None, mode=default, algorithm_name=None, algorithm_key=None, **kwargs):
        """ Create cipher with given symmetric algorithm.
        Note: default mode is CBC with initial vector composed of as many zeros as needed by algorithm.
        """
        from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
        
        if algorithm is None and algorithm_name is not None:
            if hasattr(algorithms, algorithm_name):
                algorithm = getattr(algorithms, algorithm_name)(algorithm_key)
            else:
                raise TechnicalException(f"Algorithm '{algorithm_name}' is not managed")
        if mode is default:
            mode = modes.CBC(b'\x00' * (algorithm.block_size // 8))
        
        return Cipher(algorithm, mode, **kwargs)
        
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import copy


class InspectContext(object):
    """ Base inspect context
    In usual use cases, the appropriate sub-class should be instantiated.
    """
    
    __instance_default = None
    
    def __init__(self, find_type=None, inspect_context=None):
        if find_type is not None:
            self.__find_type = find_type
        elif inspect_context is not None:
            self.__find_type = inspect_context.find_type
        else:
            self.__find_type = None
    
    @property
    def find_type(self):
        """
        @return Find type
        """
        return self.__find_type
    
    def with_find_type(self, find_type):
        """
        Note: if wanted find type is the same, self instance is returned, else a new one is returned
        @param find_type Find type
        @return Same context but with given find type
        """
        if find_type == self.find_type:
            return self
        else:
            res = copy.deepcopy(self)
            res.find_type = find_type
            return res
    
    @staticmethod
    def default(find_type=None):
        """
        @return Default inspect context
        """
        if InspectContext.__instance_default is None:
            InspectContext.__instance_default = InspectContext()
        res = InspectContext.__instance_default
        
        if find_type is not None:
            res = res.with_find_type(find_type)
        
        return res
    
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2023 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import csv
import json
from holado_core.common.exceptions.technical_exception import TechnicalException
import itertools
from holado_core.common.exceptions.functional_exception import FunctionalException
import os
from holado_core.common.tools.converters.converter import Converter
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)



class ScenarioDurationManager(object):
    def __init__(self):
        self.__execution_historic = None
    
    def import_execution_historic(self, filepath):
        with open(filepath, "r") as fin:
            self.__execution_historic = json.load(fin)
        
    def create_file_scenario_duration_limits(self, filepath, scenario_duration_limits):
        header = ["Duration Limit", "Nb of Scenarios", "Total Spent Time", "Nb of Scenarios (limit)", "Total Spent Time (limit)", "Mean Duration (limit)"]
        with open(filepath, "w", newline='') as fout:
            dw = csv.DictWriter(fout, fieldnames=header, delimiter=';')#, quoting=csv.QUOTE_NONE)
            dw.writeheader()
            
            for sd in scenario_duration_limits:
                dw.writerow(sd)
        
    def compute_scenario_duration_limits(self, duration_limits=None):
        if duration_limits is None:
            duration_limits = sorted(set(itertools.chain(*((0.1*x, x, 10*x, 60*x, 60*60*x) for x in range(1,10))))) + [60*60*24*30]

        res = [self.__compute_scenario_duration_limit(dur_lim) for dur_lim in duration_limits]
        
        # Compute base data in limit: spent time, nb of scenario
        res[0]["Total Spent Time (limit)"] = res[0]["Total Spent Time"]
        res[0]["Nb of Scenarios (limit)"] = res[0]["Nb of Scenarios"]
        for ind in range(1,len(res)):
            res[ind]["Total Spent Time (limit)"] = res[ind]["Total Spent Time"] - res[ind-1]["Total Spent Time"]
            res[ind]["Nb of Scenarios (limit)"] = res[ind]["Nb of Scenarios"] - res[ind-1]["Nb of Scenarios"]
        
        # Compute mean duration in limit
        for ind in range(len(res)):
            if res[ind]["Nb of Scenarios (limit)"] > 0:
                res[ind]["Mean Duration (limit)"] = res[ind]["Total Spent Time (limit)"] / res[ind]["Nb of Scenarios (limit)"]
            else:
                res[ind]["Mean Duration (limit)"] = None
        
        return res
    
    def __compute_scenario_duration_limit(self, duration_limit):
        res = {"Duration Limit" : duration_limit,
               "Nb of Scenarios" : 0,
               "Total Spent Time" : 0}
        for eh_feat in self.__execution_historic:
            for eh_sce in eh_feat['scenarios']:
                if eh_sce['scenario']['status'] == 'passed':
                    sce_dur = eh_sce['scenario']['duration']
                    if sce_dur < duration_limit:
                        res["Nb of Scenarios"] += 1
                        res["Total Spent Time"] += sce_dur
        return res
        
    def create_file_scenario_duration_tags(self, filepath, scenario_duration_tags):
        header = ["Feature", "Scenario", "File", "Line", "Status", "Duration", "Current tag", "New tag"]
        if scenario_duration_tags and len(scenario_duration_tags) > 0:
            header.extend(set(scenario_duration_tags[0].keys()) - set(header))
            
        with open(filepath, "w", newline='') as fout:
            dw = csv.DictWriter(fout, fieldnames=header, delimiter=';')#, quoting=csv.QUOTE_NONE)
            dw.writeheader()
            
            for sd in scenario_duration_tags:
                dw.writerow(sd)
    
    def compute_scenario_duration_tags(self, duration_limit_tags, default_tag, tag_prefix="ScenarioDuration=", missing_tag=True, new_tag=True, unchanged_tag=False, with_failed=False, add_apply=False):
        if duration_limit_tags is None or not isinstance(duration_limit_tags, list):
            raise TechnicalException("Parameter 'duration_limit_tags' must be specified as a list of tuples (limit, tag)")
        if default_tag is None:
            raise TechnicalException("Parameter 'default_tag' must be specified")
            
        res = []
        for eh_feat in self.__execution_historic:
            for eh_sce in eh_feat['scenarios']:
                sce_res = {"Feature" : eh_feat['feature']['name'], 
                           "Scenario" : eh_sce['scenario']['name'], 
                           "File" : eh_sce['scenario']['filename'], 
                           "Line" : eh_sce['scenario']['line'], 
                           "Status" : eh_sce['scenario']['status'], 
                           "Duration" : eh_sce['scenario']['duration'], 
                           "Current tag" : None, 
                           "New tag" : None}
                
                cur_tag = None
                for tag in eh_sce['scenario']['tags']:
                    tag = str(tag)
                    if tag.startswith(tag_prefix):
                        cur_tag = tag[len(tag_prefix):]
                        break
                sce_res["Current tag"] = cur_tag
                    
                cur_status = eh_sce['scenario']['status']
                new_tag = None
                if cur_status == "passed":
                    sce_dur = float(eh_sce['scenario']['duration'])
                    for lim_tag in duration_limit_tags:
                        if sce_dur < lim_tag[0]:
                            new_tag = lim_tag[1]
                            break
                    if new_tag is None:
                        new_tag = default_tag
                    sce_res["New tag"] = new_tag
                    
                if add_apply:
                    apply = None
                    if new_tag is not None:
                        if cur_tag is None:
                            apply = True
                        elif new_tag != cur_tag:
                            if cur_tag.endswith("!fixed"):
                                apply = False
                            else:
                                apply = True
                    sce_res["Apply"] = apply
                    
                if (with_failed and cur_status != "passed" 
                    or cur_status == "passed" 
                        and (missing_tag and cur_tag is None)):
                    res.append(sce_res)
        return res
    
    def update_scenario_durations(self, scenario_duration_tags_filepath, features_path, tag_prefix="ScenarioDuration=", dry_run=False):
        sdt_by_file = {}
        with open(scenario_duration_tags_filepath, 'r', newline='') as csvfile:
            reader = csv.DictReader(csvfile, delimiter=';')
            
            # Verify presence of 'Apply' column
            if 'Apply' not in reader.fieldnames:
                raise FunctionalException(f"File '{scenario_duration_tags_filepath}' doesn't have column 'Apply'. Please regenerate it with column 'Apply'.")
            
            for row in reader:
                file = row["File"]
                if file not in sdt_by_file:
                    sdt_by_file[file] = []
                sdt_by_file[file].append(row)
                
        logger.info(f"Updating scenario duration tags with prefix '{tag_prefix}' for scenarios in folder '{features_path}'")
        for file in sorted(sdt_by_file.keys()):
            sorted_sdt = sorted(sdt_by_file[file], key=lambda x: -int(x["Line"]))
            for sdt in sorted_sdt:
                self.__update_scenario_duration(sdt, features_path, tag_prefix, dry_run)
        
    def __update_scenario_duration(self, scenario_duration_tag, features_path, tag_prefix, dry_run):
        apply = scenario_duration_tag["Apply"]
        if apply and Converter.is_boolean(apply) and Converter.to_boolean(apply):
            file = scenario_duration_tag["File"]
            line_nb = int(scenario_duration_tag["Line"])
            cur_tag = scenario_duration_tag["Current tag"]
            new_tag = scenario_duration_tag["New tag"]
            
            if file.startswith('features'):
                file = file[len('features')+1:]
            filepath = os.path.join(features_path, file)
            
            logger.info(f"    in {file} l.{line_nb}: '{cur_tag}' -> '{new_tag}'")
            self.__update_scenario_duration_in_file(filepath, line_nb, cur_tag, new_tag, tag_prefix, dry_run)
        
    def __update_scenario_duration_in_file(self, filepath, line_nb, cur_tag, new_tag, tag_prefix, dry_run):
        # Read file content
        with open(filepath, 'r') as fin:
            lines = fin.readlines()
        
        # Verify line line_nb is a scenario
        if not lines[line_nb].strip().startswith("Scenario:"):
            raise FunctionalException(f"In file '{filepath}', line {line_nb} is not a scenario. This happens when scenario duration tags file is not synchronized with feature files.")
        
        # Find current tag in file
        index_duration = None
        index = line_nb - 2
        while index >= 0:
            line_content = lines[index].strip()
            if line_content.startswith("@" + tag_prefix):
                index_duration = index
                break
            elif not self.__is_line_before_scenario(line_content):
                break
            else:
                index += 1
                
        # Extract and verify current tag
        if index_duration is not None:
            scenario_tag = line_content[len("@" + tag_prefix):]
            if scenario_tag != cur_tag:
                logger.warning(f"        WARNING: Actual tag in file is '{scenario_tag}' (expected: '{cur_tag}')")
        elif cur_tag:
            logger.warning(f"        WARNING: Scenario has no actual tag in file (expected: '{cur_tag}')")
            
        # Compute new meta line
        index_sce = lines[line_nb - 1].find("Scenario")
        new_line = " " * index_sce + "@" + tag_prefix + new_tag + "\n"
        
        # Apply modification
        if index_duration is not None:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"{' '*8}Replacing line {index_duration+1}:\n{' '*12}old: [{lines[index_duration][:-1]}]\n{' '*12}new: [{new_line[:-1]}]")
            if not dry_run:
                lines[index_duration] = new_line
        else:
            index_insert = line_nb - 1
            while True:
                prev_content = lines[index_insert - 1].strip() if index_insert > 0 else None
                if prev_content is not None and len(prev_content) > 2 and prev_content[1].isupper() and prev_content > new_line.strip():
                    index_insert -= 1
                else:
                    break
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"{' '*8}Insert in line {index_insert+1}: [{new_line[:-1]}]")
            if not dry_run:
                lines.insert(index_insert, new_line)
        
        # Save in file 
        if not dry_run:
            with open(filepath, 'w') as fout:
                fout.writelines(lines)
        
    def __is_line_before_scenario(self, line):
        return not any((line.strip().startswith(txt) for txt in ['Given', 'When', 'Then', 'Feature', '|', '"""', "'''"]))
        
        


# -*- coding: utf-8 -*-

import os
import sys
import logging



# Add testing solution project source path
here = os.path.abspath(os.path.dirname(__file__))
source_path = os.path.normpath(os.path.join(here, 'src'))
sys.path.insert(0, source_path)

# Add HolAdo source paths (needed when using a clone of HolAdo project)
from initialize_holado import insert_holado_source_paths  # @UnresolvedImport
insert_holado_source_paths()


# Configure HolAdo
import holado
from context.session_context import TSSessionContext  # @UnresolvedImport
# holado.initialize(TSessionContext=TSSessionContext, 
holado.initialize(TSessionContext=None, 
                  logging_config_file_path=os.path.join(here, 'logging.conf'), log_level=logging.INFO,
                  garbage_collector_periodicity=None)


# Import generic environment methods
from behave_environment import *  # @UnresolvedImport

# Define project specific environment methods
# TestConfig.profile_memory_in_features = True
# TestConfig.profile_memory_in_scenarios = True






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.inspectors.tools.inspect_context import InspectContext
from holado_core.common.inspectors.tools.inspect_parameters import InspectParameters


class InspectBuilder(object):
    """ Inspect builder, used to create new InspectContext and InspectParameters.
    
    It is used to store available modules and finder types with Inspector.default_inspect_builder,
    and activated modules and finder types with Inspector.inspect_builder (or for GUI, GUIWindow.inspect_builder).
    By default all available modules are activated  it can be modified by overriding inspect_builder property in a new sub-class.
    """
    
    def __init__(self):
        self.__inspector = None
        
        self.__module_names = []
        self.__inspect_builder_by_module_name = {}
        
        self.__default_context = None
        self.__default_parameters = None
    
    def initialize(self, inspector):
        """
        @param inspector Inspector
        """
        self.__inspector = inspector
    
    @property
    def inspector(self):
        return self.__inspector
        
    @property
    def module_names(self):
        """
        @return Module names in order of priority
        """
        return self.__module_names
    
    def add_module(self, module_name, module_inspect_builder = None):
        """
        Add given module in list of managed modules, with given inspect builder
        @param module_name Module name
        @param module_inspect_builder Inspect builder to use with given module
        """
        if module_name in self.__module_names:
            raise TechnicalException(f"Module '{module_name}' already exists")
            
        if module_inspect_builder is None:
            module_inspector = self.__inspector.get_inspector_for_module(module_name)
            module_inspect_builder = module_inspector.default_inspect_builder
            
        self.__module_names.append(module_name)
        self.__inspect_builder_by_module_name[module_name] = module_inspect_builder
    
    def remove_module(self, module_name):
        """
        Remove given module from managed modules
        @param module_name Module name
        """
        if module_name not in self.__module_names:
            raise TechnicalException(f"Module '{module_name}' doesn't exist")
        self.__module_names.remove(module_name)
        self.__inspect_builder_by_module_name.remove(module_name)
    
    def remove_all_modules(self):
        """
        Remove all modules from managed modules
        """
        self.__module_names.clear()
        self.__inspect_builder_by_module_name.clear()
    
    def get_inspect_builder_for_module(self, module_name):
        """
        @param module_name Module name
        @return Inspect builder for given module name
        """
        if module_name not in self.__inspect_builder_by_module_name:
            raise TechnicalException(f"Module '{module_name}' doesn't exist")
        return self.__inspect_builder_by_module_name[module_name]
    
    def get_parameters_for_module(self, module_name, inspect_parameters):
        """
        @param module_name Module name
        @param inspect_parameters Current inspect parameters
        @return Module inspect parameters updated with current inspect parameter
        """
        return self.get_inspect_builder_for_module(module_name).parameters().get_module_parameters(inspect_parameters)
    
    @property
    def default_context(self):
        """
        Get or create default context.
        This method is usually used to build the default context.
        @return Default inspect context
        """
        if self.__default_context is None:
            self.__default_context = self._initialize_default_context()
        return self.__default_context
    
    def _initialize_default_context(self):
        return InspectContext.default()
        
    def context(self, inspect_context=None):
        """
        @param inspect_context Inspect context to update
        @return Updated inspect context
        """
        res = inspect_context
        if res is None:
            res = self.default_context
        return res
    
    @property
    def default_parameters(self):
        """
        Get or create default parameters.
        This method is usually used to build the default parameters.
        @return Default inspect parameters
        """
        if self.__default_parameters is None:
            self.__default_parameters = self.initialize_default_parameters()
        return self.__default_parameters
    
    def initialize_default_parameters(self):
        return InspectParameters.default_with_raise()
        
    def parameters(self, inspect_parameters=None, raise_exception=None):
        """
        @param inspect_parameters Inspect parameters to update
        @param raise_exception Whether raise exceptions or not
        @return Find parameters
        """
        res = inspect_parameters
        if res is None:
            res = self.default_parameters
            
        if raise_exception is not None:
            res = res.with_raise(raise_exception)
            
        return res
    
    def parameters_with_raise(self):
        """
        @return Default inspect parameters with raises
        """
        return self.parameters(True)
    
    def parameters_without_raise(self):
        """
        @return Default inspect parameters without any raise
        """
        return self.parameters(False)
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.tools import Tools
from holado.common.handlers.undefined import undefined_value


logger = logging.getLogger(__name__)

try:
    import redis  # @UnresolvedImport
    with_redis = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"RedisClient is not available. Initialization failed on error: {exc}")
    with_redis = False


class RedisClient(object):
    """
    Redis client
    
    This class adds high level commands to a Redis client.
    All native commands (https://redis.io/docs/latest/commands/) are callable through "internal_client" property.
    """
    
    @classmethod
    def is_available(cls):
        return with_redis
    
    def __init__(self, name, **kwargs):
        self.__name = name
        self.__kwargs = kwargs

        self.__client = self.__new_client(**kwargs)
        
        # Verify server responds to ping
        try:
            self.ping_server()
        except Exception as exc:
            raise FunctionalException(f"Failed to ping Redis server with parameters: {self.__kwargs}") from exc
        
    @property
    def name(self):
        return self.__name
    
    @property
    def host(self):
        if 'host' in self.__kwargs:
            return self.__kwargs['host']
        else:
            raise TechnicalException(f"'host' is not in client parameters (defined parameters: {self.__kwargs})")
        
    @property
    def port(self):
        if 'port' in self.__kwargs:
            return self.__kwargs['port']
        else:
            raise TechnicalException(f"'port' is not in client parameters (defined parameters: {self.__kwargs})")
        
    def __new_client(self, **kwargs):
        redis_kwargs = dict(kwargs)
        if 'credential_provider.type' in redis_kwargs:
            cp_type = redis_kwargs.pop('credential_provider.type')
            if cp_type == 'UsernamePassword':
                username = redis_kwargs.pop('credential_provider.username')
                pwd = redis_kwargs.pop('credential_provider.password')
                creds_provider = redis.UsernamePasswordCredentialProvider(username, pwd)
                redis_kwargs['credential_provider'] = creds_provider
            else:
                raise TechnicalException(f"Unexpected credential provider type '{cp_type}' (possible types: 'UsernamePassword')")
            
        return redis.Redis(**redis_kwargs)
    
    if with_redis:
        @property    
        def internal_client(self) -> redis.Redis:
            return self.__client
    
    def ping_server(self, raise_exception=True):
        """
        Ping server.
        If raise_excpetion is True, raise an exception rather than returning False.
        """
        res = self.internal_client.ping()
        if not res and raise_exception:
            raise FunctionalException(f"Redis server {self.host}:{self.port} doesn't respond to ping")
        return res
        
    def exist_key(self, key):
        nb = self.internal_client.exists(key)
        return nb > 0
        
    def get_keys(self, glob_pattern=None, count=undefined_value):
        """
        Get keys of given pattern.
        WARNING: This method should not be used if the number of keys can be huge, otherwise it can take a long time.
                 In this case, it is recommended to change redis use in order to avoid any key pattern.
        @param count: if count is undefined_value (default), internal client 'keys' method is used, else internal client 'scan' method is used
        """
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Getting keys with pattern '{glob_pattern}'...")
        
        if count is undefined_value:
            res = self.internal_client.keys(pattern=glob_pattern)
        else:
            res = []
            cursor = 0
            while True:
                cursor, cur_res = self.internal_client.scan(cursor=cursor, match=glob_pattern, count=count)
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Getting keys with pattern '{glob_pattern}': add {len(cur_res)} keys ; cursor: {cursor}")
                res.extend(cur_res)
                if cursor == 0:
                    break
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Getting keys with pattern '{glob_pattern}' -> {len(res)} keys: {res}")
        return res
        
    def delete_keys_matching(self, glob_pattern=None, recursive=False, raise_exception=True, do_unlink=True):
        """
        Delete keys of given pattern.
        WARNING: This method should not be used if the number of keys can be huge, otherwise it can take a long time.
                 In this case, it is recommended to use internal client 'delete' or 'unlink' method.
        @param do_unlink: If try, internal client 'unlink' method is used instead of 'delete' (default: True)
        """
        keys = self.get_keys(glob_pattern)
        while len(keys) > 0:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Deleting {len(keys)} keys: {keys}")
            for key in keys:
                self.delete_key(key, raise_exception=False, do_unlink=do_unlink)
    
            # Verify delete succeeded
            old_keys = keys
            keys = self.get_keys(glob_pattern)
            not_deleted_keys = set(keys).intersection(set(old_keys))
            if not recursive and len(not_deleted_keys) > 0:
                msg = f"{len(keys)} keys are not deleted and still present: {keys}"
                if raise_exception:
                    raise TechnicalException(msg)
                else:
                    logger.warning(msg)
    
            # Manage recursive
            if not recursive:
                break
            
    def delete_key(self, key, check_exist_before=False, raise_exception=True, do_unlink=True):
        """
        Delete a key.
        If check_exist_before is True and key doesn't exist, it returns immediately.
        @param do_unlink: If True, internal client 'unlink' method is used instead of 'delete' (default: True)
        """
        if check_exist_before and not self.exist_key(key):
            return
        
        if do_unlink:
            result = self.internal_client.unlink(key)
        else:
            result = self.internal_client.delete(key)
        if result == 0:
            if self.exist_key(key):
                msg = f"Failed to delete key '{key}'."
                if raise_exception:
                    raise TechnicalException(msg)
                else:
                    logger.warning(msg)
            else:
                msg = f"Failed to delete key '{key}', it doesn't exist."
                logger.warning(msg)
        else:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Deleted key '{key}'")
            
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import os
from glob import glob
import shutil
import re
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
from datetime import datetime
from holado_core.common.tools.tools import Tools
from pathlib import Path
from holado_python.common.tools.datetime import DateTime
from holado.holado_config import Config

logger = logging.getLogger(__name__)


class PathManager(object):

    def __init__(self):
        pass
    
    def get_user_group(self, path):
        p = Path(path)
        return p.owner(), p.group()
    
    def get_mode(self, path):
        return os.stat(path).st_mode & 0o777
    
    def makedirs(self, path, mode = None, is_directory = False, user = None, group = None):
        if not is_directory:
            path = os.path.dirname(path)
        if not os.path.exists(path):
            if mode:
                os.makedirs(path, mode)
            else:
                os.makedirs(path)
            self.chown(path, user, group)
            
    def makefile(self, path, mode = None, user = None, group = None):
        if not os.path.exists(path):
            self.makedirs(path, mode, False, user, group)
            open(path, 'a').close()
            
            if mode:
                os.chmod(path, mode)
            self.chown(path, user, group)
            
    def remove_path(self, path, ignore_errors=False):
        if os.path.isfile(path):
            try:
                os.remove(path)
            except Exception as exc:
                if not ignore_errors:
                    raise exc
        elif os.path.isdir(path):
            shutil.rmtree(path, ignore_errors=ignore_errors)
            
    def remove_paths(self, glob_pattern, ignore_errors=False):
        paths = glob(glob_pattern)
        for path in paths:
            self.remove_path(path, ignore_errors)
                
    def copy_path_recursively(self, src_path, dst_path, filter_patterns = None, ignore_patterns = None, do_log = False, log_prefix = "", user = None, group = None):
    #         logging.debug("Copying path '{}' -> '{}' (ignoring {})".format(src_path, dst_path, ignore_patterns))
        # Copy path
        if os.path.isfile(src_path):
            if self.__filter_path(src_path, filter_patterns=filter_patterns, ignore_patterns=ignore_patterns, do_log=do_log, log_prefix=log_prefix):
                if do_log:
                    logger.debug("{}Copy file '{}' -> '{}".format(log_prefix, src_path, dst_path))
                self.makedirs(dst_path, user=user, group=group)
                shutil.copy2(src_path, dst_path)
                self.chown(dst_path, user, group)
        elif os.path.isdir(src_path):
            lp = os.listdir(src_path)
            for cp in lp:
                cur_src_path = os.path.join(src_path, cp)
                cur_dst_path = os.path.join(dst_path, cp)
                self.copy_path_recursively(cur_src_path, cur_dst_path, filter_patterns=filter_patterns, ignore_patterns=ignore_patterns, do_log=do_log, log_prefix=log_prefix, user=user, group=group)
    
    def __filter_paths(self, paths, filter_patterns = None, ignore_patterns = None, do_log = False, log_prefix = ""):
        if filter_patterns is not None and not isinstance(filter_patterns, list):
            raise TechnicalException("Parameter 'filter_patterns' must be None or a list of patterns")
        if ignore_patterns is not None and not isinstance(ignore_patterns, list):
            raise TechnicalException("Parameter 'ignore_patterns' must be None or a list of patterns")
        
        res = []
        for path in paths:
            if self.__filter_path(path, filter_patterns, ignore_patterns, do_log, log_prefix):
                res.append(path)
                
        return res
    
    def __filter_path(self, path, filter_patterns = None, ignore_patterns = None, do_log = False, log_prefix = ""):
        if filter_patterns is not None and not isinstance(filter_patterns, list):
            raise TechnicalException("Parameter 'filter_patterns' must be None or a list of patterns")
        if ignore_patterns is not None and not isinstance(ignore_patterns, list):
            raise TechnicalException("Parameter 'ignore_patterns' must be None or a list of patterns")
        
        if filter_patterns is not None and len(filter_patterns) > 0:
            is_matching_a_filter = False
            for pattern in filter_patterns:
                if re.match(pattern, path):
                    if do_log:
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.debug(f"{log_prefix}File '{path}' is matching filter '{pattern}'")
                    is_matching_a_filter = True
                    break
            if not is_matching_a_filter:
                if do_log:
                    if Tools.do_log(logger, logging.DEBUG):
                        logger.debug(f"{log_prefix}File '{path}' is excluded by filters {filter_patterns}")
                return False

        if ignore_patterns is not None and len(ignore_patterns) > 0:
            is_ignored = False
            for pattern in ignore_patterns:
                if re.match(pattern, path):
                    if do_log:
                        if Tools.do_log(logger, logging.DEBUG):
                            logger.debug(f"{log_prefix}File '{path}' is ignored since it matches pattern '{pattern}'")
                    is_ignored = True
                    break
            if is_ignored:
                return False

        return True
    
    def rename(self, src, dst, raise_if_exists=True):
        if raise_if_exists:
            os.rename(src, dst)
        else:
            os.replace(src, dst)
    
    def get_user_home_path(self):
        import pathlib
        return str(pathlib.Path.home())
    
    def get_local_resources_path(self, name=None):
        base_path = os.getenv('HOLADO_LOCAL_RESOURCES_BASEDIR')
        if base_path is None:
            base_path = os.path.join(self.get_user_home_path(), '.holado', 'resources')
        
        if name is not None:
            return os.path.join(base_path, name)
        else:
            return base_path
    
    def get_reports_path(self, name=None, with_application_group=True):
        base_path = os.getenv('HOLADO_OUTPUT_BASEDIR')
        if base_path is None:
            base_path = os.path.join(self.get_user_home_path(), '.holado', 'output')
        
        res = os.path.join(base_path, "reports")
        if with_application_group and Config.application_group is not None:
            res = os.path.join(res, Config.application_group)
        if name is not None:
            res = os.path.join(res, name)
        return res
    
    def chown(self, path, user = None, group = None):
        if user is not None or group is not None:
            if user is None or group is None:
                raise TechnicalException(f"User and group (name or ID) cannot be None (user={user} ; group={group})")
            shutil.chown(path, user, group)
    
    def check_file_exists(self, path, do_exist=True, raise_exception=True):
        """
        @param path Path
        @param do_exist Define to check if file exists or not
        @param raise_exception If True, raises an exception rather than returning False
        @return True if given path is an existing file
        """
        res = (do_exist == os.path.exists(path))
        
        if not res and raise_exception:
            raise FunctionalException(f"File '{path}' " + "doesn't exist" if do_exist else "exists")
        return res
    
    def get_timestamped_path(self, prefix, ext, dt=None, dt_format="%Y%m%d-%H%M%S"):
        ext = ext.strip('.')
        if dt is None:
            dt = DateTime.now()
        now_str = datetime.strftime(dt, dt_format)
        return f"{prefix}_{now_str}.{ext}"
    
    def find_files(self, dir_path, prefix=None, subdir_relative_path=None, since_datetime=None):
        res = []
        if os.path.exists(dir_path):
            for filename in os.listdir(dir_path):
                file_path = os.path.join(dir_path, filename)
                
                if subdir_relative_path is not None:
                    # Filter directories that doesn't contain expected sub-file
                    if not os.path.isdir(file_path):
                        continue
                    file_path = os.path.join(file_path, subdir_relative_path)
                    if not os.path.exists(file_path):
                        continue
                else:
                    # Filter paths that are not files
                    if not os.path.isfile(file_path):
                        continue
                
                # Filter on prefix
                if prefix is not None and not filename.startswith(prefix):
                    continue
                
                # Filter on file modification datetime
                if since_datetime is not None:
                    dt_last_modif = DateTime.timestamp_to_datetime(os.path.getmtime(file_path))
                    if dt_last_modif < since_datetime:
                        continue
                
                # File is matching criteria
                res.append(file_path)
        return res
    
    def find_file(self, dir_path, prefix=None, since_datetime=None):
        files = self.find_files(dir_path, prefix=prefix, since_datetime=since_datetime)
        
        if len(files) == 0:
            raise FunctionalException(f"Unable to find a file starting with '{prefix}' in folder '{dir_path}'")
        elif len(files) > 1:
            raise FunctionalException(f"Found many ({len(files)}) files starting with '{prefix}' in folder '{dir_path}': {files}")
        return files[0]
    
